# Copyright 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ironic_python_agent import encoding


class RESTError(Exception, encoding.Serializable):
    """Base class for errors generated in ironic-python-client."""
    # NOTE(JoshNang) `message` should not end with a period
    message = 'An error occurred'
    details = 'An unexpected error occurred. Please try back later.'
    status_code = 500
    serializable_fields = ('type', 'code', 'message', 'details')

    def __init__(self, details=None, *args, **kwargs):
        super(RESTError, self).__init__(*args, **kwargs)
        self.type = self.__class__.__name__
        self.code = self.status_code
        if details:
            self.details = details

    def __str__(self):
        return "{}: {}".format(self.message, self.details)

    def __repr__(self):
        """Should look like RESTError('message: details')"""
        return "{}('{}')".format(self.__class__.__name__, self.__str__())


class InvalidContentError(RESTError):
    """Error which occurs when a user supplies invalid content.

    Either because that content cannot be parsed according to the advertised
    `Content-Type`, or due to a content validation error.
    """
    message = 'Invalid request body'
    status_code = 400

    def __init__(self, details):
        super(InvalidContentError, self).__init__(details)


class NotFound(RESTError):
    """Error which occurs if a non-existent API endpoint is called."""
    message = 'Not found'
    status_code = 404
    details = 'The requested URL was not found.'


class CommandExecutionError(RESTError):
    """Error raised when a command fails to execute."""

    message = 'Command execution failed'

    def __init__(self, details):
        super(CommandExecutionError, self).__init__(details)


class InvalidCommandError(InvalidContentError):
    """Error which is raised when an unknown command is issued."""

    message = 'Invalid command'

    def __init__(self, details):
        super(InvalidCommandError, self).__init__(details)


class InvalidCommandParamsError(InvalidContentError):
    """Error which is raised when command parameters are invalid."""

    message = 'Invalid command parameters'

    def __init__(self, details):
        super(InvalidCommandParamsError, self).__init__(details)


class RequestedObjectNotFoundError(NotFound):
    def __init__(self, type_descr, obj_id):
        details = '{} with id {} not found.'.format(type_descr, obj_id)
        super(RequestedObjectNotFoundError, self).__init__(details)


class AgentIsBusy(CommandExecutionError):

    message = 'Agent is busy'
    status_code = 409

    def __init__(self, command_name):
        super().__init__('executing command %s' % command_name)


class IronicAPIError(RESTError):
    """Error raised when a call to the agent API fails."""

    message = 'Error in call to ironic-api'

    def __init__(self, details):
        super(IronicAPIError, self).__init__(details)


class HeartbeatError(IronicAPIError):
    """Error raised when a heartbeat to the agent API fails."""

    message = 'Error heartbeating to agent API'

    def __init__(self, details):
        super(HeartbeatError, self).__init__(details)


class HeartbeatConflictError(IronicAPIError):
    """ConflictError raised when a heartbeat to the agent API fails."""

    message = 'ConflictError heartbeating to agent API'

    def __init__(self, details):
        super(HeartbeatConflictError, self).__init__(details)


class LookupNodeError(IronicAPIError):
    """Error raised when the node lookup to the Ironic API fails."""

    message = 'Error getting configuration from Ironic'

    def __init__(self, details):
        super(LookupNodeError, self).__init__(details)


class LookupAgentIPError(IronicAPIError):
    """Error raised when automatic IP lookup fails."""

    message = 'Error finding IP for Ironic Agent'

    def __init__(self, details):
        super(LookupAgentIPError, self).__init__(details)


class ImageDownloadError(RESTError):
    """Error raised when an image cannot be downloaded."""

    message = 'Error downloading image'

    def __init__(self, image_id, msg):
        details = 'Download of image {} failed: {}'.format(image_id, msg)
        self.secondary_message = msg
        super(ImageDownloadError, self).__init__(details)


class ImageChecksumError(RESTError):
    """Error raised when an image fails to verify against its checksum."""

    message = 'Error verifying image checksum'
    details_str = ('Image failed to verify against checksum. location: {}; '
                   'image ID: {}; image checksum: {}; verification '
                   'checksum: {}')

    def __init__(self, image_id, image_location, checksum,
                 calculated_checksum):
        details = self.details_str.format(image_location, image_id, checksum,
                                          calculated_checksum)
        super(ImageChecksumError, self).__init__(details)


class ImageWriteError(RESTError):
    """Error raised when an image cannot be written to a device."""

    message = 'Error writing image to device'

    def __init__(self, device, exit_code, stdout, stderr):
        details = ('Writing image to device {} failed with exit code '
                   '{}. stdout: {}. stderr: {}')
        details = details.format(device, exit_code, stdout, stderr)
        super(ImageWriteError, self).__init__(details)


class SystemRebootError(RESTError):
    """Error raised when a system cannot reboot."""

    message = 'Error rebooting system'

    def __init__(self, exit_code, stdout, stderr):
        details = ('Reboot script failed with exit code {}. stdout: '
                   '{}. stderr: {}.')
        details = details.format(exit_code, stdout, stderr)
        super(SystemRebootError, self).__init__(details)


class BlockDeviceEraseError(RESTError):
    """Error raised when an error occurs erasing a block device."""

    message = 'Error erasing block device'

    def __init__(self, details):
        super(BlockDeviceEraseError, self).__init__(details)


class BlockDeviceError(RESTError):
    """Error raised when a block devices causes an unknown error."""

    message = 'Block device caused unknown error'

    def __init__(self, details):
        super(BlockDeviceError, self).__init__(details)


class SoftwareRAIDError(RESTError):
    """Error raised when a Software RAID causes an error."""

    message = 'Software RAID caused unknown error'

    def __init__(self, details):
        super(SoftwareRAIDError, self).__init__(details)


class VirtualMediaBootError(RESTError):
    """Error raised when virtual media device cannot be found for config."""

    message = 'Configuring agent from virtual media failed'

    def __init__(self, details):
        super(VirtualMediaBootError, self).__init__(details)


class ExtensionError(RESTError):
    pass


class UnknownNodeError(RESTError):
    """Error raised when the agent is not associated with an Ironic node."""

    message = 'Agent is not associated with an Ironic node'

    def __init__(self, details=None):
        super(UnknownNodeError, self).__init__(details)


class HardwareManagerNotFound(RESTError):
    """Error raised when no valid HardwareManager can be found."""

    message = 'No valid HardwareManager found'

    def __init__(self, details=None):
        super(HardwareManagerNotFound, self).__init__(details)


class HardwareManagerMethodNotFound(RESTError):
    """Error raised when all HardwareManagers fail to handle a method."""

    message = 'No HardwareManager found to handle method'

    def __init__(self, method):
        details = 'Could not find method: {}'.format(method)
        super(HardwareManagerMethodNotFound, self).__init__(details)


class IncompatibleHardwareMethodError(RESTError):
    """Error raised when HardwareManager method incompatible with hardware."""

    message = 'HardwareManager method is not compatible with hardware'

    def __init__(self, details=None):
        super(IncompatibleHardwareMethodError, self).__init__(details)


class VersionMismatch(RESTError):
    """Error raised when Ironic and the Agent have different versions.

    If the agent version has changed since get_clean_steps or get_deploy_steps
    was called by the Ironic conductor, it indicates the agent has been updated
    (either on purpose, or a new agent was deployed and the node was rebooted).
    Since we cannot know if the upgraded IPA will work with cleaning/deploy as
    it stands (steps could have different priorities, either in IPA or in
    other Ironic interfaces), we should restart the process from the start.

    """
    message = (
        'Hardware managers version mismatch, reload agent with correct version'
    )

    def __init__(self, agent_version, node_version):
        self.status_code = 409
        details = ('Current versions: {}, versions used by ironic: {}'
                   .format(agent_version, node_version))
        super(VersionMismatch, self).__init__(details)


class CleaningError(RESTError):
    """Error raised when a cleaning step fails."""

    message = 'Clean step failed'

    def __init__(self, details=None):
        super(CleaningError, self).__init__(details)


class DeploymentError(RESTError):
    """Error raised when a deploy step fails."""

    message = 'Deploy step failed'

    def __init__(self, details=None):
        super(DeploymentError, self).__init__(details)


class IncompatibleNumaFormatError(RESTError):
    """Error raised when unexpected format data in NUMA node."""

    message = 'Error in NUMA node data format'


class DeviceNotFound(NotFound):
    """Error raised when the device to deploy the image onto is not found."""

    message = ('Error finding the disk or partition device to deploy '
               'the image onto')

    def __init__(self, details):
        super(DeviceNotFound, self).__init__(details)


# This is not something we return to a user, so we don't inherit it from
# RESTError.
class InspectionError(Exception):
    """Failure during inspection."""


class ClockSyncError(RESTError):
    """Error raised when attempting to sync the system clock."""

    message = 'Error syncing system clock'


class HeartbeatConnectionError(IronicAPIError):
    """Transitory connection failure occured attempting to contact the API."""

    message = ("Error attempting to heartbeat - Possible transitory network "
               "failure or blocking port may be present.")

    def __init__(self, details):
        super(HeartbeatConnectionError, self).__init__(details)



# Copyright 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from oslo_config import cfg
from oslo_log import log
from oslo_service import sslutils
from oslo_utils import strutils

from ironic_python_agent import agent
from ironic_python_agent import config
from ironic_python_agent import utils

CONF = cfg.CONF


def run():
    """Entrypoint for IronicPythonAgent."""
    # NOTE(dtantsur): this must happen very early of the files from
    # /etc/ironic-python-agent.d won't be loaded
    utils.copy_config_from_vmedia()

    log.register_options(CONF)
    CONF(args=sys.argv[1:])
    # Debug option comes from oslo.log, allow overriding it via kernel cmdline
    ipa_debug = config.APARAMS.get('ipa-debug')
    if ipa_debug is not None:
        ipa_debug = strutils.bool_from_string(ipa_debug)
        CONF.set_override('debug', ipa_debug)
    log.setup(CONF, 'ironic-python-agent')
    # Used for TLS configuration
    sslutils.register_opts(CONF)

    logger = log.getLogger(__name__)
    logger.debug("Configuration:")
    CONF.log_opt_values(logger, log.DEBUG)
    utils.log_early_log_to_logger()
    agent.IronicPythonAgent(CONF.api_url,
                            agent.Host(hostname=CONF.advertise_host,
                                       port=CONF.advertise_port),
                            agent.Host(hostname=CONF.listen_host,
                                       port=CONF.listen_port),
                            CONF.ip_lookup_attempts,
                            CONF.ip_lookup_sleep,
                            CONF.network_interface,
                            CONF.lookup_timeout,
                            CONF.lookup_interval,
                            False,
                            CONF.agent_token,
                            CONF.hardware_initialization_delay,
                            CONF.advertise_protocol).run()



# Copyright 2015 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

from ironic_lib import mdns
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import excutils
import requests
import stevedore
import tenacity

from ironic_python_agent import config
from ironic_python_agent import encoding
from ironic_python_agent import errors
from ironic_python_agent import hardware
from ironic_python_agent import utils


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

_DHCP_RETRY_INTERVAL = 2
_COLLECTOR_NS = 'ironic_python_agent.inspector.collectors'
_NO_LOGGING_FIELDS = ('logs',)


def _extension_manager_err_callback(names):
    raise errors.InspectionError('Failed to load collector %s' % names)


def extension_manager(names):
    return stevedore.NamedExtensionManager(
        _COLLECTOR_NS, names=names, name_order=True,
        on_missing_entrypoints_callback=_extension_manager_err_callback)


def _get_collector_names():
    return [x.strip() for x in CONF.inspection_collectors.split(',')
            if x.strip()]


def inspect():
    """Optionally run inspection on the current node.

    If ``inspection_callback_url`` is set in the configuration, get
    the hardware inventory from the node and post it back to the inspector.

    :return: node UUID if inspection was successful, None if associated node
             was not found in inspector cache. None is also returned if
             inspector support is not enabled.
    """
    if not CONF.inspection_callback_url:
        LOG.info('Inspection is disabled, skipping')
        return

    if CONF.inspection_callback_url == 'mdns':
        LOG.debug('Fetching the inspection URL from mDNS')
        url, params = mdns.get_endpoint('baremetal-introspection')
        # We expect a proper catalog URL, which doesn't include any path.
        CONF.set_override('inspection_callback_url',
                          url.rstrip('/') + '/v1/continue')
        config.override(params)

    collector_names = _get_collector_names()
    LOG.info('inspection is enabled with collectors %s', collector_names)

    # NOTE(dtantsur): inspection process tries to delay raising any exceptions
    # until after we posted some data back to inspector. This is because
    # inspection is run automatically on (mostly) unknown nodes, so if it
    # fails, we don't have much information for debugging.
    failures = utils.AccumulatedFailures(exc_class=errors.InspectionError)
    data = {}

    try:
        ext_mgr = extension_manager(collector_names)
        collectors = [(ext.name, ext.plugin) for ext in ext_mgr]
    except Exception as exc:
        with excutils.save_and_reraise_exception():
            failures.add(exc)
            call_inspector(data, failures)

    for name, collector in collectors:
        try:
            collector(data, failures)
        except Exception as exc:
            # No reraise here, try to keep going
            failures.add('collector %s failed: %s', name, exc)

    resp = call_inspector(data, failures)

    # Now raise everything we were delaying
    failures.raise_if_needed()

    if resp is None:
        raise errors.InspectionError('stopping inspection, as inspector '
                                     'returned an error')

    LOG.info('inspection finished successfully')
    return resp.get('uuid')


_RETRY_WAIT = 5
_RETRY_ATTEMPTS = 5


def call_inspector(data, failures):
    """Post data to inspector."""
    data['error'] = failures.get_error()

    LOG.info('posting collected data to %s', CONF.inspection_callback_url)
    LOG.debug('collected data: %s',
              {k: v for k, v in data.items() if k not in _NO_LOGGING_FIELDS})

    encoder = encoding.RESTJSONEncoder()
    data = encoder.encode(data)
    verify, cert = utils.get_ssl_client_options(CONF)

    @tenacity.retry(
        retry=tenacity.retry_if_exception_type(
            requests.exceptions.ConnectionError),
        stop=tenacity.stop_after_attempt(_RETRY_ATTEMPTS),
        wait=tenacity.wait_fixed(_RETRY_WAIT),
        reraise=True)
    def _post_to_inspector():
        return requests.post(CONF.inspection_callback_url, data=data,
                             verify=verify, cert=cert)

    resp = _post_to_inspector()
    if resp.status_code >= 400:
        LOG.error('inspector %s error %d: %s, proceeding with lookup',
                  CONF.inspection_callback_url,
                  resp.status_code, resp.content.decode('utf-8'))
        return

    return resp.json()


def _normalize_mac(mac):
    """Convert MAC to a well-known format aa:bb:cc:dd:ee:ff."""
    if '-' in mac:
        # pxelinux format is 01-aa-bb-cc-dd-ee-ff
        mac = mac.split('-', 1)[1]
        mac = mac.replace('-', ':')
    return mac.lower()


def wait_for_dhcp():
    """Wait until NIC's get their IP addresses via DHCP or timeout happens.

    Depending on the value of inspection_dhcp_all_interfaces configuration
    option will wait for either all or only PXE booting NIC.

    Note: only supports IPv4 addresses for now.

    :return: True if all NIC's got IP addresses, False if timeout happened.
             Also returns True if waiting is disabled via configuration.
    """
    if not CONF.inspection_dhcp_wait_timeout:
        return True

    pxe_mac = utils.get_agent_params().get('BOOTIF')
    if pxe_mac:
        pxe_mac = _normalize_mac(pxe_mac)
    elif not CONF.inspection_dhcp_all_interfaces:
        LOG.warning('No PXE boot interface known - not waiting for it '
                    'to get the IP address')
        return False

    threshold = time.time() + CONF.inspection_dhcp_wait_timeout
    while time.time() <= threshold:
        interfaces = hardware.dispatch_to_managers('list_network_interfaces')
        interfaces = [iface for iface in interfaces
                      if CONF.inspection_dhcp_all_interfaces
                      or iface.mac_address.lower() == pxe_mac]
        missing = [iface.name for iface in interfaces
                   if not iface.ipv4_address]
        if not missing:
            return True

        LOG.debug('Still waiting for interfaces %s to get IP addresses',
                  missing)
        time.sleep(_DHCP_RETRY_INTERVAL)

    LOG.warning('Not all network interfaces received IP addresses in '
                '%(timeout)d seconds: %(missing)s',
                {'timeout': CONF.inspection_dhcp_wait_timeout,
                 'missing': missing})
    return False


def collect_default(data, failures):
    """The default inspection collector.

    This is the only collector that is called by default. It collects
    the whole inventory as returned by the hardware manager(s).

    It also tries to get BMC address, PXE boot device and the expected
    root device.

    :param data: mutable data that we'll send to inspector
    :param failures: AccumulatedFailures object
    """
    wait_for_dhcp()
    inventory = hardware.list_hardware_info()

    data['inventory'] = inventory
    # Replicate the same logic as in deploy. We need to make sure that when
    # root device hints are not set, inspector will use the same root disk as
    # will be used for deploy.
    try:
        root_disk = utils.guess_root_disk(inventory['disks'][:])
    except errors.DeviceNotFound:
        root_disk = None
        LOG.warning('no suitable root device detected')
    else:
        data['root_disk'] = root_disk
        LOG.debug('default root device is %s', root_disk.name)
    # The boot interface might not be present, we don't count it as failure.
    # TODO(dtantsur): stop using the boot_interface field.
    data['boot_interface'] = inventory['boot'].pxe_interface
    LOG.debug('boot devices was %s', data['boot_interface'])
    LOG.debug('BMC IP address: %s', inventory.get('bmc_address'))
    data['configuration'] = {
        'collectors': _get_collector_names(),
        'managers': [mgr.get_version() for mgr in hardware.get_managers()],
    }


def collect_logs(data, failures):
    """Collect system logs from the ramdisk.

    As inspection runs before any nodes details are known, it's handy to have
    logs returned with data. This collector sends logs to inspector in format
    expected by the 'ramdisk_error' plugin: base64 encoded tar.gz.

    This collector should be installed last in the collector chain, otherwise
    it won't collect enough logs.

    This collector does not report failures.

    :param data: mutable data that we'll send to inspector
    :param failures: AccumulatedFailures object
    """
    try:
        data['logs'] = utils.collect_system_logs(journald_max_lines=10000)
    except errors.CommandExecutionError:
        LOG.warning('failed to get system journal')
        return


def collect_extra_hardware(data, failures):
    """Collect detailed inventory using 'hardware-detect' utility.

    Recognizes ipa-inspection-benchmarks with list of benchmarks (possible
    values are cpu, disk, mem) to run. No benchmarks are run by default, as
    they're pretty time-consuming.

    Puts collected data as JSON under 'data' key.
    Requires 'hardware' python package to be installed on the ramdisk in
    addition to the packages in requirements.txt.

    :param data: mutable data that we'll send to inspector
    :param failures: AccumulatedFailures object
    """
    benchmarks = utils.get_agent_params().get('ipa-inspection-benchmarks', [])
    if benchmarks:
        benchmarks = ['--benchmark'] + benchmarks.split(',')

    try:
        out, err = utils.execute('hardware-detect', *benchmarks)
    except (processutils.ProcessExecutionError, OSError) as exc:
        failures.add('failed to run hardware-detect utility: %s', exc)
        return

    try:
        data['data'] = jsonutils.loads(out)
    except ValueError as exc:
        msg = 'JSON returned from hardware-detect cannot be decoded: %s'
        failures.add(msg, exc)


def collect_pci_devices_info(data, failures):
    """Collect a list of PCI devices.

    Each PCI device entry in list is a dictionary containing vendor_id and
    product_id keys, which will be then used by the ironic inspector to
    distinguish various PCI devices.

    The data is gathered from /sys/bus/pci/devices directory.

    :param data: mutable data that we'll send to inspector
    :param failures: AccumulatedFailures object
    """
    pci_devices_path = '/sys/bus/pci/devices'
    pci_devices_info = []
    try:
        subdirs = os.listdir(pci_devices_path)
    except OSError as exc:
        msg = 'Failed to get list of PCI devices: %s'
        failures.add(msg, exc)
        return
    for subdir in subdirs:
        if not os.path.isdir(os.path.join(pci_devices_path, subdir)):
            continue
        try:
            # note(sborkows): ids located in files inside PCI devices
            # directory are stored in hex format (0x1234 for example) and
            # we only need that part after 'x' delimiter
            with open(os.path.join(pci_devices_path, subdir,
                                   'vendor')) as vendor_file:
                vendor = vendor_file.read().strip().split('x')[1]
            with open(os.path.join(pci_devices_path, subdir,
                                   'device')) as vendor_device:
                device = vendor_device.read().strip().split('x')[1]
            with open(os.path.join(pci_devices_path, subdir,
                                   'class')) as vendor_device:
                pci_class = vendor_device.read().strip().split('x')[1]
        except IOError as exc:
            LOG.warning('Failed to gather vendor id, product id or pci class '
                        'from PCI device %s: %s', subdir, exc)
            continue
        except IndexError as exc:
            LOG.warning('Wrong format of vendor id, product id or pci class '
                        'in PCI device %s: %s', subdir, exc)
            continue

        pci_revision = None
        pci_revision_path = os.path.join(pci_devices_path, subdir,
                                         'revision')
        if os.path.isfile(pci_revision_path):
            try:
                with open(pci_revision_path) as revision_file:
                    pci_revision = revision_file.read().strip().split('x')[1]
            except IOError as exc:
                LOG.warning('Failed to gather PCI revision from PCI '
                            'device %s: %s', subdir, exc)
            except IndexError as exc:
                LOG.warning('Wrong format of PCI revision in PCI '
                            'device %s: %s', subdir, exc)

        LOG.debug(
            'Found a PCI device with vendor id %s, product id %s, class %s '
            'and revision %s', vendor, device, pci_class, pci_revision)
        pci_devices_info.append({'vendor_id': vendor,
                                 'product_id': device,
                                 'class': pci_class,
                                 'revision': pci_revision,
                                 'bus': subdir})
    data['pci_devices'] = pci_devices_info



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of the inject_files deploy step."""

import base64
import contextlib
import os

from ironic_lib import disk_utils
from ironic_lib import utils as ironic_utils
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log

from ironic_python_agent import errors
from ironic_python_agent import hardware
from ironic_python_agent import utils


CONF = cfg.CONF
LOG = log.getLogger(__name__)


ARGSINFO = {
    "files": {
        "description": (
            "Files to inject, a list of file structures with keys: 'path' "
            "(path to the file), 'partition' (partition specifier), "
            "'content' (base64 encoded string), 'mode' (new file mode) and "
            "'dirmode' (mode for the leaf directory, if created). "
            "Merged with the values from node.properties[inject_files]."
        ),
        "required": False,
    },
    "verify_ca": {
        "description": (
            "Whether to verify TLS certificates. Global agent options "
            "are used by default."
        ),
        "required": False,
    }
}


def inject_files(node, ports, files, verify_ca=True):
    """A deploy step to inject arbitrary files.

    :param node: A dictionary of the node object
    :param ports: A list of dictionaries containing information
                  of ports for the node
    :param files: See ARGSINFO.
    :param verify_ca: Whether to verify TLS certificate.
    :raises: InvalidCommandParamsError
    """
    files = _validate_files(
        node['properties'].get('inject_files') or [],
        files or [])
    if not files:
        LOG.info('No files to inject')
        return

    http_get = utils.StreamingClient(verify_ca)
    root_dev = hardware.dispatch_to_managers('get_os_install_device')

    for fl in files:
        _inject_one(node, ports, fl, root_dev, http_get)


def _inject_one(node, ports, fl, root_dev, http_get):
    """Inject one file.

    :param node: A dictionary of the node object
    :param ports: A list of dictionaries containing information
                  of ports for the node
    :param fl: File information.
    :param root_dev: Root device used for the current node.
    :param http_get: Context manager to get HTTP URLs.
    """
    with _find_and_mount_path(fl['path'], fl.get('partition'),
                              root_dev) as path:
        if fl.get('deleted'):
            ironic_utils.unlink_without_raise(path)
            return

        try:
            dirpath = os.path.dirname(path)
            try:
                os.makedirs(dirpath)
            except FileExistsError:
                pass
            else:
                # Use chmod here and below to avoid relying on umask
                if fl.get('dirmode'):
                    os.chmod(dirpath, fl['dirmode'])

            content = fl['content']
            with open(path, 'wb') as fp:
                if '://' in content:
                    # Allow node-specific URLs to be used in a deploy template
                    url = content.format(node=node, ports=ports)
                    with http_get(url) as resp:
                        for chunk in resp:
                            fp.write(chunk)
                else:
                    fp.write(base64.b64decode(content))

            if fl.get('mode'):
                os.chmod(path, fl['mode'])

            if fl.get('owner') is not None or fl.get('group') is not None:
                # -1 means do not change
                os.chown(path, fl.get('owner', -1), fl.get('group', -1))
        except Exception as exc:
            LOG.exception('Failed to process file %s', fl)
            raise errors.CommandExecutionError(
                'Failed to process file %s. %s: %s'
                % (fl, type(exc).__class__, exc))


@contextlib.contextmanager
def _find_and_mount_path(path, partition, root_dev):
    """Find the specified path on a device.

    Tries to find the suitable device for the file based on the ``path`` and
    ``partition``, mount the device and provides the actual full path.

    :param path: Path to the file to find.
    :param partition: Device to find the file on or None.
    :param root_dev: Root device from the hardware manager.
    :return: Context manager that yields the full path to the file.
    """
    path = os.path.normpath(path.strip('/'))  # to make path joining work
    if partition:
        try:
            part_num = int(partition)
        except ValueError:
            with ironic_utils.mounted(partition) as part_path:
                yield os.path.join(part_path, path)
        else:
            # TODO(dtantsur): switch to ironic-lib instead:
            # https://review.opendev.org/c/openstack/ironic-lib/+/774502
            part_template = '%s%s'
            if 'nvme' in root_dev:
                part_template = '%sp%s'
            part_dev = part_template % (root_dev, part_num)

            with ironic_utils.mounted(part_dev) as part_path:
                yield os.path.join(part_path, path)
    else:
        try:
            # This turns e.g. etc/sysctl.d/my.conf into etc + sysctl.d/my.conf
            detect_dir, rest_dir = path.split('/', 1)
        except ValueError:
            # Validation ensures that files in / have "partition" present,
            # checking here just in case.
            raise errors.InvalidCommandParamsError(
                "Invalid path %s, must be an absolute path to a file" % path)

        with find_partition_with_path(detect_dir, root_dev) as part_path:
            yield os.path.join(part_path, rest_dir)


@contextlib.contextmanager
def find_partition_with_path(path, device=None):
    """Find a partition with the given path.

    :param path: Expected path.
    :param device: Target device. If None, the root device is used.
    :returns: A context manager that will unmount and delete the temporary
        mount point on exit.
    """
    if device is None:
        device = hardware.dispatch_to_managers('get_os_install_device')
    partitions = disk_utils.list_partitions(device)
    # Make os.path.join work as expected
    lookup_path = path.lstrip('/')

    for part in partitions:
        if 'lvm' in part['flags']:
            LOG.debug('Skipping LVM partition %s', part)
            continue

        # TODO(dtantsur): switch to ironic-lib instead:
        # https://review.opendev.org/c/openstack/ironic-lib/+/774502
        part_template = '%s%s'
        if 'nvme' in device:
            part_template = '%sp%s'
        part_path = part_template % (device, part['number'])

        LOG.debug('Inspecting partition %s for path %s', part, path)
        try:
            with ironic_utils.mounted(part_path) as local_path:
                found_path = os.path.join(local_path, lookup_path)
                if not os.path.isdir(found_path):
                    continue

                LOG.info('Path %s has been found on partition %s', path, part)
                yield found_path
                return
        except processutils.ProcessExecutionError as exc:
            LOG.warning('Failure when inspecting partition %s: %s', part, exc)

    raise errors.DeviceNotFound("No partition found with path %s, scanned: %s"
                                % (path, partitions))


def _validate_files(from_properties, from_args):
    """Sanity check for files."""
    if not isinstance(from_properties, list):
        raise errors.InvalidCommandParamsError(
            "The `inject_files` node property must be a list, got %s"
            % type(from_properties).__name__)
    if not isinstance(from_args, list):
        raise errors.InvalidCommandParamsError(
            "The `files` argument must be a list, got %s"
            % type(from_args).__name__)

    files = from_properties + from_args
    failures = []

    for fl in files:
        unknown = set(fl) - {'path', 'partition', 'content', 'deleted', 'mode',
                             'dirmode', 'owner', 'group'}
        if unknown:
            failures.append('unexpected fields in %s: %s'
                            % (fl, ', '.join(unknown)))

        if not fl.get('path'):
            failures.append('expected a path in %s' % fl)
        elif os.path.dirname(fl['path']) == '/' and not fl.get('partition'):
            failures.append('%s in root directory requires "partition"' % fl)
        elif fl['path'].endswith('/'):
            failures.append('directories not supported for %s' % fl)

        if fl.get('content') and fl.get('deleted'):
            failures.append('content cannot be used with deleted in %s' % fl)

        for field in ('owner', 'group', 'mode', 'dirmode'):
            if field in fl and type(fl[field]) is not int:
                failures.append('%s must be a number in %s' % (field, fl))

    if failures:
        raise errors.InvalidCommandParamsError(
            "Validation of files failed: %s" % '; '.join(failures))

    return files



# Copyright 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log

from ironic_python_agent import hardware

LOG = log.getLogger()


# All the helper methods should be kept outside of the HardwareManager
# so they'll never get accidentally called by dispatch_to_managers()
def _initialize_hardware():
    """Example method for initalizing hardware."""
    # Perform any operations here that are required to initialize your
    # hardware.
    LOG.debug('Loading drivers, settling udevs, and generally initalizing')
    pass


def _detect_hardware():
    """Example method for hardware detection."""
    # For this example, return true if hardware is detected, false if not
    LOG.debug('Looking for example device')
    return True


def _is_latest_firmware():
    """Detect if device is running latest firmware."""
    # Actually detect the firmware version instead of returning here.
    return True


def _upgrade_firmware():
    """Upgrade firmware on device."""
    # Actually perform firmware upgrade instead of returning here.
    return True


class ExampleDeviceHardwareManager(hardware.HardwareManager):
    """Example hardware manager to support a single device"""

    # All hardware managers have a name and a version.
    # Version should be bumped anytime a change is introduced. This will
    # signal to Ironic that if automatic node cleaning is in progress to
    # restart it from the beginning, to ensure consistency. The value can
    # be anything; it's checked for equality against previously seen
    # name:manager pairs.
    HARDWARE_MANAGER_NAME = 'ExampleDeviceHardwareManager'
    HARDWARE_MANAGER_VERSION = '1'

    def evaluate_hardware_support(self):
        """Declare level of hardware support provided.

        Since this example covers a case of supporting a specific device,
        this method is where you would do anything needed to initalize that
        device, including loading drivers, and then detect if one exists.

        In some cases, if you expect the hardware to be available on any node
        running this hardware manager, or it's undetectable, you may want to
        return a static value here.

        Be aware all managers' loaded in IPA will run this method before IPA
        performs a lookup or begins heartbeating, so the time needed to
        execute this method will make cleaning and deploying slower.

        :returns: HardwareSupport level for this manager.
        """
        _initialize_hardware()
        if _detect_hardware():
            # This actually resolves down to an int. Upstream IPA will never
            # return a value higher than 2 (HardwareSupport.MAINLINE). This
            # means your managers should always be SERVICE_PROVIDER or higher.
            LOG.debug('Found example device, returning SERVICE_PROVIDER')
            return hardware.HardwareSupport.SERVICE_PROVIDER
        else:
            # If the hardware isn't supported, return HardwareSupport.NONE (0)
            # in order to prevent IPA from loading its clean steps or
            # attempting to use any methods inside it.
            LOG.debug('No example devices found, returning NONE')
            return hardware.HardwareSupport.NONE

    def get_clean_steps(self, node, ports):
        """Get a list of clean steps with priority.

        Define any clean steps added by this manager here. These will be mixed
        with other loaded managers that support this hardware, and ordered by
        priority. Higher priority steps run earlier.

        Note that out-of-band clean steps may also be provided by Ironic.
        These will follow the same priority ordering even though they are not
        executed by IPA.

        There is *no guarantee whatsoever* that steps defined here will be
        executed by this HardwareManager. When it comes time to run these
        steps, they'll be called using dispatch_to_managers() just like any
        other IPA HardwareManager method. This means if they are unique to
        your hardware, they should be uniquely named. For example,
        upgrade_firmware would be a bad step name. Whereas
        upgrade_foobar_device_firmware would be better.

        :param node: The node object as provided by Ironic.
        :param ports: Port objects as provided by Ironic.
        :returns: A list of cleaning steps, as a list of dicts.
        """
        # While obviously you could actively run code here, generally this
        # should just return a static value, as any initialization and
        # detection should've been done in evaluate_hardware_support().
        return [{
            'step': 'upgrade_example_device_model1234_firmware',
            'priority': 37,
            # If you need Ironic to coordinate a reboot after this step
            # runs, but before continuing cleaning, this should be true.
            'reboot_requested': True,
            # If it's safe for Ironic to abort cleaning while this step
            # runs, this should be true.
            'abortable': False
        }]

    def upgrade_example_device_model1234_firmware(self, node, ports):
        """Upgrade firmware on Example Device Model #1234."""
        # Any commands needed to perform the firmware upgrade should go here.
        # If you plan on actually flashing firmware every cleaning cycle, you
        # should ensure your device will not experience flash exhaustion. A
        # good practice in some environments would be to check the firmware
        # version against a constant in the code, and noop the method if an
        # upgrade is not needed.
        if _is_latest_firmware():
            LOG.debug('Latest firmware already flashed, skipping')
            # Return values are ignored here on success
            return True
        else:
            LOG.debug('Firmware version X found, upgrading to Y')
            # Perform firmware upgrade.
            try:
                _upgrade_firmware()
            except Exception as e:
                # Log and pass through the exception so cleaning will fail
                LOG.exception(e)
                raise
            return True



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from ironic_python_agent import errors
from ironic_python_agent import numa_inspector as numa_insp
from ironic_python_agent.tests.unit import base
from ironic_python_agent import utils


class TestCollectNumaTopologyInfo(base.IronicAgentTest):
    def setUp(self):
        super(TestCollectNumaTopologyInfo, self).setUp()
        self.data = {}
        self.failures = utils.AccumulatedFailures()

    @mock.patch.object(numa_insp, 'get_nodes_nics_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_cores_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_memory_info', autospec=True)
    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    def test_collect_success(self, mock_listdir, mock_isdir, mock_memory_info,
                             mock_cores_info, mock_nics_info):
        numa_node_dirs = ['node0', 'node1']
        mock_listdir.return_value = numa_node_dirs
        mock_isdir.return_value = True
        mock_memory_info.return_value = [{'numa_node': 0, 'size_kb': 1560000},
                                         {'numa_node': 1, 'size_kb': 1200000}]
        mock_cores_info.return_value = [{'cpu': 0, 'numa_node': 0,
                                         'thread_siblings': [0, 1, 2, 3]},
                                        {'cpu': 1, 'numa_node': 0,
                                         'thread_siblings': [4, 5, 6]},
                                        {'cpu': 0, 'numa_node': 1,
                                         'thread_siblings': [16, 17]},
                                        {'cpu': 1, 'numa_node': 1,
                                         'thread_siblings': [18, 19]}]
        mock_nics_info.return_value = [{'name': 'enp0s01', 'numa_node': 0},
                                       {'name': 'enp0s02', 'numa_node': 1}]
        expected_numa_info = {"ram": [{'numa_node': 0, 'size_kb': 1560000},
                                      {'numa_node': 1, 'size_kb': 1200000}],
                              "cpus": [{'cpu': 0, 'numa_node': 0,
                                        'thread_siblings': [0, 1, 2, 3]},
                                       {'cpu': 1, 'numa_node': 0,
                                        'thread_siblings': [4, 5, 6]},
                                       {'cpu': 0, 'numa_node': 1,
                                        'thread_siblings': [16, 17]},
                                       {'cpu': 1, 'numa_node': 1,
                                        'thread_siblings': [18, 19]}],
                              "nics": [{'name': 'enp0s01', 'numa_node': 0},
                                       {'name': 'enp0s02', 'numa_node': 1}]}
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            numa_insp.collect_numa_topology_info(self.data, self.failures)
        self.assertEqual(expected_numa_info, self.data["numa_topology"])

    @mock.patch.object(os.path, 'isdir', autospec=True)
    def test_collect_no_numa_dirs(self, mock_isdir):
        mock_isdir.return_value = False
        numa_insp.collect_numa_topology_info(self.data, self.failures)
        self.assertNotIn("numa_topology", self.data)

    @mock.patch.object(numa_insp, 'get_nodes_nics_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_cores_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_memory_info', autospec=True)
    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    def test_collect_no_nics_dirs(self, mock_listdir, mock_isdir,
                                  mock_memory_info, mock_cores_info,
                                  mock_nics_info):
        numa_node_dirs = ['node0', 'node1']
        mock_listdir.return_value = numa_node_dirs
        mock_isdir.return_value = True
        mock_memory_info.return_value = [{'numa_node': 0, 'size_kb': 1560000},
                                         {'numa_node': 1, 'size_kb': 1200000}]
        mock_cores_info.return_value = [{'cpu': 0, 'numa_node': 0,
                                         'thread_siblings': [0, 1, 2, 3]},
                                        {'cpu': 1, 'numa_node': 1,
                                         'thread_siblings': [4, 5, 6]}]
        mock_nics_info.side_effect = errors.IncompatibleNumaFormatError("")
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            numa_insp.collect_numa_topology_info(self.data, self.failures)
        self.assertNotIn("numa_topology", self.data)

    @mock.patch.object(numa_insp, 'get_nodes_nics_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_cores_info', autospec=True)
    @mock.patch.object(numa_insp, 'get_nodes_memory_info', autospec=True)
    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    def test_collect_failure(self, mock_listdir, mock_isdir, mock_memory_info,
                             mock_cores_info, mock_nics_info):
        numa_node_dirs = ['node0', 'node1']
        mock_listdir.return_value = numa_node_dirs
        mock_isdir.return_value = True
        mock_memory_info.side_effect = errors.IncompatibleNumaFormatError("")
        mock_cores_info.side_effect = errors.IncompatibleNumaFormatError("")
        mock_nics_info.side_effect = errors.IncompatibleNumaFormatError("")
        numa_insp.collect_numa_topology_info(self.data, self.failures)
        self.assertNotIn("numa_topology", self.data)
        self.assertFalse(self.failures)


class TestGetNumaTopologyInfo(base.IronicAgentTest):
    def setUp(self):
        super(TestGetNumaTopologyInfo, self).setUp()
        self.data = {}
        self.failures = utils.AccumulatedFailures()

    def test_get_numa_node_id_valid_format(self):
        numa_node_dir = '/sys/devices/system/node/node0'
        expected_numa_node_id = 0
        numa_node_id = numa_insp.get_numa_node_id(numa_node_dir)
        self.assertEqual(expected_numa_node_id, numa_node_id)

    def test_get_numa_node_id_invalid_format(self):
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_numa_node_id,
                          '/sys/devices/system/node/node-*0')
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_numa_node_id,
                          '/sys/devices/system/node/nod')
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_numa_node_id,
                          '')

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_get_nodes_memory_info(self, mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/node0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = [0, 1]
        reads = [['Node 0 Ignored Line',
                  'Node 0 MemTotal: 1560000 kB'],
                 ['Node 1 MemTotal: 1200000 kB',
                  'Node 1 Ignored Line']]
        expected_meminfo = [{'numa_node': 0, 'size_kb': 1560000},
                            {'numa_node': 1, 'size_kb': 1200000}]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_meminfo_file = mock.MagicMock()
            mock_meminfo_file.__enter__.side_effect = reads
            mock_open.return_value = mock_meminfo_file
            ram = numa_insp.get_nodes_memory_info(numa_node_dirs)
        self.assertListEqual(expected_meminfo, ram)

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_bad_nodes_memory_info(self, mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/node0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = [0, 1]
        reads = [['Node 0 MemTotal: 1560000 kB'], IOError]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_meminfo_file = mock.MagicMock()
            mock_meminfo_file.__enter__.side_effect = reads
            mock_open.return_value = mock_meminfo_file
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_memory_info,
                              numa_node_dirs)

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_nodes_invalid_numa_format_memory_info(self, mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/node0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = [0, 1]
        reads = [['Node 0: MemTotal: 1560000 kB'],
                 ['Node 1 MemTotal: 1200000 kB']]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_meminfo_file = mock.MagicMock()
            mock_meminfo_file.__enter__.side_effect = reads
            mock_open.return_value = mock_meminfo_file
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_memory_info,
                              numa_node_dirs)

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_nodes_invalid_memory_unit(self, mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/node0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = [0, 1]
        reads = [['Node 0 MemTotal: 1560000 TB'],
                 ['Node 1 MemTotal: 1200000 kB']]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_meminfo_file = mock.MagicMock()
            mock_meminfo_file.__enter__.side_effect = reads
            mock_open.return_value = mock_meminfo_file
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_memory_info,
                              numa_node_dirs)

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_get_numa_node_id_invalid_format_memory_info(self,
                                                         mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/node-*0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = errors.IncompatibleNumaFormatError
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_nodes_memory_info,
                          numa_node_dirs)

    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_get_nodes_cores_info(self, mock_node_id,
                                  mock_listdir, mock_isdir):
        numa_node_dirs = ['/sys/devices/system/node/node0',
                          '/sys/devices/system/node/node1']
        mock_node_id.side_effect = [0, 1]
        mock_listdir.side_effect = [['cpu0', 'cpu1', 'cpu2',
                                     'cpu3', 'cpu4'],
                                    ['cpu5', 'cpu6', 'cpu7']]
        mock_isdir.return_value = True
        reads = ['0', '0', '1', '1', '1', '0', '0', '0']
        expected_cores_info = [{'cpu': 0, 'numa_node': 0,
                                'thread_siblings': [0, 1]},
                               {'cpu': 1, 'numa_node': 0,
                                'thread_siblings': [2, 3, 4]},
                               {'cpu': 0, 'numa_node': 1,
                                'thread_siblings': [5, 6, 7]}]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_core_id_file = mock_open.return_value.read
            mock_core_id_file.side_effect = reads
            cpus = numa_insp.get_nodes_cores_info(numa_node_dirs)
        self.assertEqual(len(cpus), len(expected_cores_info))
        for cpu in cpus:
            self.assertIn(cpu, expected_cores_info)

    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_bad_nodes_cores_info(self, mock_node_id,
                                  mock_listdir, mock_isdir):
        numa_node_dirs = ['/sys/devices/system/node/node0']
        mock_node_id.return_value = 0
        thread_dirs = ['cpu0', 'cpu1', 'cpu2', 'cpu3', 'cpu4', 'cpu5']
        mock_listdir.return_value = thread_dirs
        mock_isdir.return_value = True
        reads = ['0', '0', '1', '1', '1', IOError]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_core_id_file = mock_open.return_value.read
            mock_core_id_file.side_effect = reads
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_cores_info,
                              numa_node_dirs)

    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_get_numa_node_id_invalid_format_cores_info(self,
                                                        mock_node_id):
        numa_node_dirs = ['/sys/devices/system/node/nodeid0']
        mock_node_id.side_effect = errors.IncompatibleNumaFormatError
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_nodes_cores_info,
                          numa_node_dirs)

    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_nodes_invalid_threaddir_format_cores_info(self, mock_node_id,
                                                       mock_listdir,
                                                       mock_isdir):
        numa_node_dirs = ['/sys/devices/system/node/node0']
        mock_node_id.return_value = 0
        thread_dirs = ['cpuid0', 'cpu1', 'cpu2', 'cpu3', 'cpu4', 'cpu5']
        mock_listdir.return_value = thread_dirs
        mock_isdir.return_value = True
        reads = ['0', '0', '1', '1', '1', '2']
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_core_id_file = mock_open.return_value.read
            mock_core_id_file.side_effect = reads
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_cores_info,
                              numa_node_dirs)

    @mock.patch.object(os.path, 'isdir', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    @mock.patch.object(numa_insp, 'get_numa_node_id', autospec=True)
    def test_bad_nodes_thread_dirs(self, mock_node_id,
                                   mock_listdir, mock_isdir):
        numa_node_dirs = ['/sys/devices/system/node/node0']
        mock_node_id.return_value = 0
        mock_listdir.side_effect = errors.IncompatibleNumaFormatError("")
        mock_isdir.return_value = True
        self.assertRaises(errors.IncompatibleNumaFormatError,
                          numa_insp.get_nodes_cores_info,
                          numa_node_dirs)

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    def test_get_nodes_nics_info(self, mock_listdir, mock_isfile):
        nic_dirs = ['enp0s01', 'enp0s02']
        mock_listdir.return_value = nic_dirs
        mock_isfile.return_value = True
        reads = ['0', '1']
        expected_nicsinfo = [{'name': 'enp0s01', 'numa_node': 0},
                             {'name': 'enp0s02', 'numa_node': 1}]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_nicsinfo_file = mock_open.return_value.read
            mock_nicsinfo_file.side_effect = reads
            nics = numa_insp.get_nodes_nics_info('/sys/class/net/')
        self.assertListEqual(expected_nicsinfo, nics)

    @mock.patch.object(os.path, 'isfile', autospec=True)
    @mock.patch.object(os, 'listdir', autospec=True)
    def test_bad_nodes_nics_info(self, mock_listdir, mock_isfile):
        nic_dirs = ['enp0s01', 'enp0s02']
        mock_listdir.return_value = nic_dirs
        mock_isfile.return_value = True
        reads = ['0', IOError]
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_nicsinfo_file = mock_open.return_value.read
            mock_nicsinfo_file.side_effect = reads
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_nics_info,
                              '/sys/class/net/')

    @mock.patch.object(os, 'listdir', autospec=True)
    @mock.patch.object(os.path, 'isdir', autospec=True)
    def test_no_nics_dir(self, mock_isdir, mock_listdir):
        mock_isdir.return_value = False
        nic_dirs = ['enp0s01', 'enp0s02']
        mock_listdir.return_value = nic_dirs
        reads = ['0', '1']
        mock_open = mock.mock_open()
        with mock.patch('builtins.open', mock_open):
            mock_nicsinfo_file = mock_open.return_value.read
            mock_nicsinfo_file.side_effect = reads
            self.assertRaises(errors.IncompatibleNumaFormatError,
                              numa_insp.get_nodes_nics_info,
                              '/sys/class/net/')



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import ipaddress
import os
import tempfile
from unittest import mock

from cryptography.hazmat import backends
from cryptography import x509

from ironic_python_agent.tests.unit import base as ironic_agent_base
from ironic_python_agent import tls_utils


class GenerateTestCase(ironic_agent_base.IronicAgentTest):

    def setUp(self):
        super().setUp()
        tempdir = tempfile.mkdtemp()
        self.crt_file = os.path.join(tempdir, 'localhost.crt')
        self.key_file = os.path.join(tempdir, 'localhost.key')

    def test__generate(self):
        result = tls_utils._generate_tls_certificate(self.crt_file,
                                                     self.key_file,
                                                     'localhost', '127.0.0.1')
        now = datetime.datetime.utcnow()
        self.assertTrue(result.startswith("-----BEGIN CERTIFICATE-----\n"),
                        result)
        self.assertTrue(result.endswith("\n-----END CERTIFICATE-----\n"),
                        result)
        self.assertTrue(os.path.exists(self.key_file))
        with open(self.crt_file, 'rt') as fp:
            self.assertEqual(result, fp.read())

        cert = x509.load_pem_x509_certificate(result.encode(),
                                              backends.default_backend())
        self.assertEqual([(x509.NameOID.COMMON_NAME, 'localhost')],
                         [(item.oid, item.value) for item in cert.subject])
        # Sanity check for validity range
        self.assertLess(cert.not_valid_before,
                        now - datetime.timedelta(seconds=1800))
        self.assertGreater(cert.not_valid_after,
                           now + datetime.timedelta(seconds=1800))
        subject_alt_name = cert.extensions.get_extension_for_oid(
            x509.ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
        self.assertTrue(subject_alt_name.critical)
        self.assertEqual(
            [ipaddress.IPv4Address('127.0.0.1')],
            subject_alt_name.value.get_values_for_type(x509.IPAddress))
        self.assertEqual(
            [], subject_alt_name.value.get_values_for_type(x509.DNSName))

    @mock.patch('ironic_python_agent.netutils.get_hostname', autospec=True)
    @mock.patch('os.makedirs', autospec=True)
    @mock.patch.object(tls_utils, '_generate_tls_certificate', autospec=True)
    def test_generate(self, mock_generate, mock_makedirs, mock_hostname):
        result = tls_utils.generate_tls_certificate('127.0.0.1')
        mock_generate.assert_called_once_with(result.path,
                                              result.private_key_path,
                                              mock_hostname.return_value,
                                              '127.0.0.1')



# Copyright 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ironic_lib import exception as il_exc
from oslo_log import log

from ironic_python_agent import errors
from ironic_python_agent.extensions import base
from ironic_python_agent import hardware

LOG = log.getLogger()


class CleanExtension(base.BaseAgentExtension):
    @base.sync_command('get_clean_steps')
    def get_clean_steps(self, node, ports):
        """Get the list of clean steps supported for the node and ports

        :param node: A dict representation of a node
        :param ports: A dict representation of ports attached to node

        :returns: A list of clean steps with keys step, priority, and
            reboot_requested
        """
        LOG.debug('Getting clean steps, called with node: %(node)s, '
                  'ports: %(ports)s', {'node': node, 'ports': ports})
        hardware.cache_node(node)
        # Results should be a dict, not a list
        candidate_steps = hardware.dispatch_to_all_managers('get_clean_steps',
                                                            node, ports)

        LOG.debug('Clean steps before deduplication: %s', candidate_steps)
        clean_steps = hardware.deduplicate_steps(candidate_steps)
        LOG.debug('Returning clean steps: %s', clean_steps)

        return {
            'clean_steps': clean_steps,
            'hardware_manager_version': hardware.get_current_versions(),
        }

    @base.async_command('execute_clean_step')
    def execute_clean_step(self, step, node, ports, clean_version=None,
                           **kwargs):
        """Execute a clean step.

        :param step: A clean step with 'step', 'priority' and 'interface' keys
        :param node: A dict representation of a node
        :param ports: A dict representation of ports attached to node
        :param clean_version: The clean version as returned by
                              hardware.get_current_versions() at the beginning
                              of cleaning/zapping
        :returns: a CommandResult object with command_result set to whatever
            the step returns.
        """
        # Ensure the agent is still the same version, or raise an exception
        LOG.debug('Executing clean step %s', step)
        hardware.cache_node(node)
        hardware.check_versions(clean_version)

        if 'step' not in step:
            msg = 'Malformed clean_step, no "step" key: %s' % step
            LOG.error(msg)
            raise ValueError(msg)
        try:
            result = hardware.dispatch_to_managers(step['step'], node, ports)
        except (errors.RESTError, il_exc.IronicException):
            LOG.exception('Error performing clean step %s', step['step'])
            raise
        except Exception as e:
            msg = ('Unexpected exception performing clean step %(step)s. '
                   '%(cls)s: %(err)s' % {'step': step['step'],
                                         'cls': e.__class__.__name__,
                                         'err': e})
            LOG.exception(msg)
            raise errors.CleaningError(msg)

        LOG.info('Clean step completed: %(step)s, result: %(result)s',
                 {'step': step, 'result': result})

        # Cast result tuples (like output of utils.execute) as lists, or
        # API throws errors
        if isinstance(result, tuple):
            result = list(result)

        # Return the step that was executed so we can dispatch
        # to the appropriate Ironic interface
        return {
            'clean_result': result,
            'clean_step': step
        }



# Copyright 2014 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log

from ironic_python_agent import errors
from ironic_python_agent.extensions import base

LOG = log.getLogger(__name__)


def _validate_exts(ext, flow=None):
    for task in flow:
        for method in task:
            ext_name, cmd = ext.split_command(method)
            if ext_name not in ext.ext_mgr.names():
                raise errors.RequestedObjectNotFoundError('Extension',
                                                          ext_name)
            ext_obj = ext.ext_mgr[ext_name].obj
            ext.check_cmd_presence(ext_obj, ext_name, cmd)


class FlowExtension(base.BaseAgentExtension, base.ExecuteCommandMixin):
    @base.async_command('start_flow', _validate_exts)
    def start_flow(self, flow=None):
        for task in flow:
            for method, params in task.items():
                LOG.info("Executing method %s for now", method)
                result = self.execute_command(method, **params)
                result.join()
                LOG.info("%s method's execution is done", method)
                if result.command_status == base.AgentCommandStatus.FAILED:
                    raise errors.CommandExecutionError(
                        "%s was failed" % method
                    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Logic related to handling partitions.

Imported from ironic-lib's disk_utils as of the following commit:
https://opendev.org/openstack/ironic-lib/commit/9fb5be348202f4854a455cd08f400ae12b99e1f2
"""

import base64
import gzip
import io
import math
import os
import shlex
import shutil
import stat
import tempfile

from ironic_lib import disk_utils
from ironic_lib import exception
from ironic_lib import utils
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import units
from oslo_utils import uuidutils
import requests

from ironic_python_agent import errors
from ironic_python_agent import hardware
from ironic_python_agent import utils as ipa_utils


LOG = log.getLogger()
CONF = cfg.CONF

MAX_CONFIG_DRIVE_SIZE_MB = 64

# Maximum disk size supported by MBR is 2TB (2 * 1024 * 1024 MB)
MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR = 2097152


def get_configdrive(configdrive, node_uuid, tempdir=None):
    """Get the information about size and location of the configdrive.

    :param configdrive: Base64 encoded Gzipped configdrive content or
        configdrive HTTP URL.
    :param node_uuid: Node's uuid. Used for logging.
    :param tempdir: temporary directory for the temporary configdrive file
    :raises: InstanceDeployFailure if it can't download or decode the
       config drive.
    :returns: A tuple with the size in MiB and path to the uncompressed
        configdrive file.

    """
    # Check if the configdrive option is a HTTP URL or the content directly
    is_url = utils.is_http_url(configdrive)
    if is_url:
        verify, cert = ipa_utils.get_ssl_client_options(CONF)
        timeout = CONF.image_download_connection_timeout
        # TODO(dtantsur): support proxy parameters from instance_info
        try:
            resp = requests.get(configdrive, verify=verify, cert=cert,
                                timeout=timeout)
        except requests.exceptions.RequestException as e:
            raise exception.InstanceDeployFailure(
                "Can't download the configdrive content for node %(node)s "
                "from '%(url)s'. Reason: %(reason)s" %
                {'node': node_uuid, 'url': configdrive, 'reason': e})

        if resp.status_code >= 400:
            raise exception.InstanceDeployFailure(
                "Can't download the configdrive content for node %(node)s "
                "from '%(url)s'. Got status code %(code)s, response "
                "body %(body)s" %
                {'node': node_uuid, 'url': configdrive,
                 'code': resp.status_code, 'body': resp.text})

        data = resp.content
    else:
        data = configdrive

    configdrive_file = tempfile.NamedTemporaryFile(delete=False,
                                                   prefix='configdrive',
                                                   dir=tempdir)

    try:
        data = io.BytesIO(base64.b64decode(data))
    except Exception as exc:
        if isinstance(data, bytes):
            LOG.debug('Config drive for node %(node)s is not base64 encoded '
                      '(%(error)s), assuming binary',
                      {'node': node_uuid, 'error': exc})
            configdrive_mb = int(math.ceil(len(data) / units.Mi))
            configdrive_file.write(data)
            configdrive_file.close()
            return (configdrive_mb, configdrive_file.name)
        else:
            configdrive_file.close()
            utils.unlink_without_raise(configdrive_file.name)

            error_msg = ('Config drive for node %(node)s is not base64 '
                         'encoded or the content is malformed. '
                         '%(cls)s: %(err)s.'
                         % {'node': node_uuid, 'err': exc,
                            'cls': type(exc).__name__})
            if is_url:
                error_msg += ' Downloaded from "%s".' % configdrive
            raise exception.InstanceDeployFailure(error_msg)

    configdrive_mb = 0
    with gzip.GzipFile('configdrive', 'rb', fileobj=data) as gunzipped:
        try:
            shutil.copyfileobj(gunzipped, configdrive_file)
        except EnvironmentError as e:
            # Delete the created file
            utils.unlink_without_raise(configdrive_file.name)
            raise exception.InstanceDeployFailure(
                'Encountered error while decompressing and writing '
                'config drive for node %(node)s. Error: %(exc)s' %
                {'node': node_uuid, 'exc': e})
        else:
            # Get the file size and convert to MiB
            configdrive_file.seek(0, os.SEEK_END)
            bytes_ = configdrive_file.tell()
            configdrive_mb = int(math.ceil(float(bytes_) / units.Mi))
        finally:
            configdrive_file.close()

        return (configdrive_mb, configdrive_file.name)


def get_labelled_partition(device_path, label, node_uuid):
    """Check and return if partition with given label exists

    :param device_path: The device path.
    :param label: Partition label
    :param node_uuid: UUID of the Node. Used for logging.
    :raises: InstanceDeployFailure, if any disk partitioning related
        commands fail.
    :returns: block device file for partition if it exists; otherwise it
              returns None.
    """
    disk_utils.partprobe(device_path)
    try:
        output, err = utils.execute('lsblk', '-Po', 'name,label', device_path,
                                    check_exit_code=[0, 1],
                                    use_standard_locale=True, run_as_root=True)

    except (processutils.UnknownArgumentError,
            processutils.ProcessExecutionError, OSError) as e:
        msg = ('Failed to retrieve partition labels on disk %(disk)s '
               'for node %(node)s. Error: %(error)s' %
               {'disk': device_path, 'node': node_uuid, 'error': e})
        LOG.error(msg)
        raise exception.InstanceDeployFailure(msg)

    found_part = None
    if output:
        for dev in utils.parse_device_tags(output):
            if dev['LABEL'].upper() == label.upper():
                if found_part:
                    found_2 = '/dev/%(part)s' % {'part': dev['NAME'].strip()}
                    found = [found_part, found_2]
                    raise exception.InstanceDeployFailure(
                        'More than one partition with label "%(label)s" '
                        'exists on device %(device)s for node %(node)s: '
                        '%(found)s.' %
                        {'label': label, 'device': device_path,
                         'node': node_uuid, 'found': ' and '.join(found)})
                found_part = '/dev/%(part)s' % {'part': dev['NAME'].strip()}

    return found_part


def work_on_disk(dev, root_mb, swap_mb, ephemeral_mb, ephemeral_format,
                 image_path, node_uuid, preserve_ephemeral=False,
                 configdrive=None, boot_option="netboot", boot_mode="bios",
                 tempdir=None, disk_label=None, cpu_arch="", conv_flags=None):
    """Create partitions and copy an image to the root partition.

    :param dev: Path for the device to work on.
    :param root_mb: Size of the root partition in megabytes.
    :param swap_mb: Size of the swap partition in megabytes.
    :param ephemeral_mb: Size of the ephemeral partition in megabytes. If 0,
        no ephemeral partition will be created.
    :param ephemeral_format: The type of file system to format the ephemeral
        partition.
    :param image_path: Path for the instance's disk image. If ``None``,
        the root partition is prepared but not populated.
    :param node_uuid: node's uuid. Used for logging.
    :param preserve_ephemeral: If True, no filesystem is written to the
        ephemeral block device, preserving whatever content it had (if the
        partition table has not changed).
    :param configdrive: Optional. Base64 encoded Gzipped configdrive content
                        or configdrive HTTP URL.
    :param boot_option: Can be "local" or "netboot". "netboot" by default.
    :param boot_mode: Can be "bios" or "uefi". "bios" by default.
    :param tempdir: A temporary directory
    :param disk_label: The disk label to be used when creating the
        partition table. Valid values are: "msdos", "gpt" or None; If None
        Ironic will figure it out according to the boot_mode parameter.
    :param cpu_arch: Architecture of the node the disk device belongs to.
        When using the default value of None, no architecture specific
        steps will be taken. This default should be used for x86_64. When
        set to ppc64*, architecture specific steps are taken for booting a
        partition image locally.
    :param conv_flags: Flags that need to be sent to the dd command, to control
        the conversion of the original file when copying to the host. It can
        contain several options separated by commas.
    :returns: a dictionary containing the following keys:
        'root uuid': UUID of root partition
        'efi system partition uuid': UUID of the uefi system partition
        (if boot mode is uefi).
        `partitions`: mapping of partition types to their device paths.
        NOTE: If key exists but value is None, it means partition doesn't
        exist.
    """
    # the only way for preserve_ephemeral to be set to true is if we are
    # rebuilding an instance with --preserve_ephemeral.
    commit = not preserve_ephemeral
    # now if we are committing the changes to disk clean first.
    if commit:
        disk_utils.destroy_disk_metadata(dev, node_uuid)

    try:
        # If requested, get the configdrive file and determine the size
        # of the configdrive partition
        configdrive_mb = 0
        configdrive_file = None
        if configdrive:
            configdrive_mb, configdrive_file = get_configdrive(
                configdrive, node_uuid, tempdir=tempdir)

        part_dict = disk_utils.make_partitions(dev,
                                               root_mb, swap_mb, ephemeral_mb,
                                               configdrive_mb, node_uuid,
                                               commit=commit,
                                               boot_option=boot_option,
                                               boot_mode=boot_mode,
                                               disk_label=disk_label,
                                               cpu_arch=cpu_arch)
        LOG.info("Successfully completed the disk device"
                 " %(dev)s partitioning for node %(node)s",
                 {'dev': dev, "node": node_uuid})

        ephemeral_part = part_dict.get('ephemeral')
        swap_part = part_dict.get('swap')
        configdrive_part = part_dict.get('configdrive')
        root_part = part_dict.get('root')

        if not disk_utils.is_block_device(root_part):
            raise exception.InstanceDeployFailure(
                "Root device '%s' not found" % root_part)

        for part in ('swap', 'ephemeral', 'configdrive',
                     'efi system partition', 'PReP Boot partition'):
            part_device = part_dict.get(part)
            LOG.debug("Checking for %(part)s device (%(dev)s) on node "
                      "%(node)s.", {'part': part, 'dev': part_device,
                                    'node': node_uuid})
            if part_device and not disk_utils.is_block_device(part_device):
                raise exception.InstanceDeployFailure(
                    "'%(partition)s' device '%(part_device)s' not found" %
                    {'partition': part, 'part_device': part_device})

        # If it's a uefi localboot, then we have created the efi system
        # partition.  Create a fat filesystem on it.
        if boot_mode == "uefi" and boot_option == "local":
            efi_system_part = part_dict.get('efi system partition')
            utils.mkfs(fs='vfat', path=efi_system_part, label='efi-part')

        if configdrive_part:
            # Copy the configdrive content to the configdrive partition
            disk_utils.dd(configdrive_file, configdrive_part,
                          conv_flags=conv_flags)
            LOG.info("Configdrive for node %(node)s successfully copied "
                     "onto partition %(partition)s",
                     {'node': node_uuid, 'partition': configdrive_part})

    finally:
        # If the configdrive was requested make sure we delete the file
        # after copying the content to the partition
        if configdrive_file:
            utils.unlink_without_raise(configdrive_file)

    if image_path is not None:
        disk_utils.populate_image(image_path, root_part, conv_flags=conv_flags)
        LOG.info("Image for %(node)s successfully populated",
                 {'node': node_uuid})
    else:
        LOG.debug("Root partition for %s was created, but not populated",
                  node_uuid)

    if swap_part:
        utils.mkfs(fs='swap', path=swap_part, label='swap1')
        LOG.info("Swap partition %(swap)s successfully formatted "
                 "for node %(node)s",
                 {'swap': swap_part, 'node': node_uuid})

    if ephemeral_part and not preserve_ephemeral:
        utils.mkfs(fs=ephemeral_format, path=ephemeral_part,
                   label="ephemeral0")
        LOG.info("Ephemeral partition %(ephemeral)s successfully "
                 "formatted for node %(node)s",
                 {'ephemeral': ephemeral_part, 'node': node_uuid})

    # Rescan device to get current status (e.g. reflect modification of mkfs)
    disk_utils.trigger_device_rescan(dev)

    uuids_to_return = {
        'root uuid': root_part,
        'efi system partition uuid': part_dict.get('efi system partition'),
    }

    if cpu_arch.startswith('ppc'):
        uuids_to_return[
            'PReP Boot partition uuid'
        ] = part_dict.get('PReP Boot partition')

    try:
        for part, part_dev in uuids_to_return.items():
            if part_dev:
                uuids_to_return[part] = disk_utils.block_uuid(part_dev)

    except processutils.ProcessExecutionError:
        with excutils.save_and_reraise_exception():
            LOG.error("Failed to detect %s", part)

    return dict(partitions=part_dict, **uuids_to_return)


def create_config_drive_partition(node_uuid, device, configdrive):
    """Create a partition for config drive

    Checks if the device is GPT or MBR partitioned and creates config drive
    partition accordingly.

    :param node_uuid: UUID of the Node.
    :param device: The device path.
    :param configdrive: Base64 encoded Gzipped configdrive content or
        configdrive HTTP URL.
    :raises: InstanceDeployFailure if config drive size exceeds maximum limit
        or if it fails to create config drive.
    """
    confdrive_file = None
    try:
        config_drive_part = get_labelled_partition(
            device, disk_utils.CONFIGDRIVE_LABEL, node_uuid)

        confdrive_mb, confdrive_file = get_configdrive(configdrive, node_uuid)
        if confdrive_mb > MAX_CONFIG_DRIVE_SIZE_MB:
            raise exception.InstanceDeployFailure(
                'Config drive size exceeds maximum limit of 64MiB. '
                'Size of the given config drive is %(size)d MiB for '
                'node %(node)s.'
                % {'size': confdrive_mb, 'node': node_uuid})

        LOG.debug("Adding config drive partition %(size)d MiB to "
                  "device: %(dev)s for node %(node)s",
                  {'dev': device, 'size': confdrive_mb, 'node': node_uuid})

        disk_utils.fix_gpt_partition(device, node_uuid)
        if config_drive_part:
            LOG.debug("Configdrive for node %(node)s exists at "
                      "%(part)s",
                      {'node': node_uuid, 'part': config_drive_part})
        else:
            part_uuid = None
            if disk_utils.get_partition_table_type(device) == 'gpt':
                part_uuid = uuidutils.generate_uuid()
                create_option = '0:-%dMB:0' % MAX_CONFIG_DRIVE_SIZE_MB
                uuid_option = '0:%s' % part_uuid
                utils.execute('sgdisk', '-n', create_option,
                              '-u', uuid_option, device,
                              run_as_root=True)
            else:
                cur_parts = set(part['number']
                                for part in disk_utils.list_partitions(device))

                # Check if the disk has 4 partitions. The MBR based disk
                # cannot have more than 4 partitions.
                # TODO(stendulker): One can use logical partitions to create
                # a config drive if there are 3 primary partitions.
                # https://bugs.launchpad.net/ironic/+bug/1561283
                try:
                    pp_count, lp_count = disk_utils.count_mbr_partitions(
                        device)
                except ValueError as e:
                    raise exception.InstanceDeployFailure(
                        'Failed to check the number of primary partitions '
                        'present on %(dev)s for node %(node)s. Error: '
                        '%(error)s' % {'dev': device, 'node': node_uuid,
                                       'error': e})
                if pp_count > 3:
                    raise exception.InstanceDeployFailure(
                        'Config drive cannot be created for node %(node)s. '
                        'Disk (%(dev)s) uses MBR partitioning and already '
                        'has %(parts)d primary partitions.'
                        % {'node': node_uuid, 'dev': device,
                           'parts': pp_count})

                # Check if disk size exceeds 2TB msdos limit
                startlimit = '-%dMiB' % MAX_CONFIG_DRIVE_SIZE_MB
                endlimit = '-0'
                if _is_disk_larger_than_max_size(device, node_uuid):
                    # Need to create a small partition at 2TB limit
                    LOG.warning("Disk size is larger than 2TB for "
                                "node %(node)s. Creating config drive "
                                "at the end of the disk %(disk)s.",
                                {'node': node_uuid, 'disk': device})
                    startlimit = (MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR
                                  - MAX_CONFIG_DRIVE_SIZE_MB - 1)
                    endlimit = MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR - 1

                utils.execute('parted', '-a', 'optimal', '-s', '--', device,
                              'mkpart', 'primary', 'fat32', startlimit,
                              endlimit, run_as_root=True)
            # Trigger device rescan
            disk_utils.trigger_device_rescan(device)

            if part_uuid is None:
                new_parts = {part['number']: part
                             for part in disk_utils.list_partitions(device)}
                new_part = set(new_parts) - set(cur_parts)
                if len(new_part) != 1:
                    raise exception.InstanceDeployFailure(
                        'Disk partitioning failed on device %(device)s. '
                        'Unable to retrieve config drive partition '
                        'information.' % {'device': device})

                config_drive_part = disk_utils.partition_index_to_path(
                    device, new_part.pop())
            else:
                try:
                    config_drive_part = get_partition(device, part_uuid)
                except errors.DeviceNotFound:
                    msg = ('Failed to create config drive on disk %(disk)s '
                           'for node %(node)s. Partition with UUID %(uuid)s '
                           'has not been found after creation.') % {
                               'disk': device, 'node': node_uuid,
                               'uuid': part_uuid}
                    LOG.error(msg)
                    raise exception.InstanceDeployFailure(msg)

            disk_utils.udev_settle()

            # NOTE(vsaienko): check that devise actually exists,
            # it is not handled by udevadm when using ISCSI, for more info see:
            # https://bugs.launchpad.net/ironic/+bug/1673731
            # Do not use 'udevadm settle --exit-if-exist' here
            LOG.debug('Waiting for the config drive partition %(part)s '
                      'on node %(node)s to be ready for writing.',
                      {'part': config_drive_part, 'node': node_uuid})
            utils.execute('test', '-e', config_drive_part, attempts=15,
                          delay_on_retry=True)

        disk_utils.dd(confdrive_file, config_drive_part)
        LOG.info("Configdrive for node %(node)s successfully "
                 "copied onto partition %(part)s",
                 {'node': node_uuid, 'part': config_drive_part})

    except (processutils.UnknownArgumentError,
            processutils.ProcessExecutionError, OSError) as e:
        msg = ('Failed to create config drive on disk %(disk)s '
               'for node %(node)s. Error: %(error)s' %
               {'disk': device, 'node': node_uuid, 'error': e})
        LOG.error(msg)
        raise exception.InstanceDeployFailure(msg)
    finally:
        # If the configdrive was requested make sure we delete the file
        # after copying the content to the partition
        if confdrive_file:
            utils.unlink_without_raise(confdrive_file)


def _is_disk_larger_than_max_size(device, node_uuid):
    """Check if total disk size exceeds 2TB msdos limit

    :param device: device path.
    :param node_uuid: node's uuid. Used for logging.
    :raises: InstanceDeployFailure, if any disk partitioning related
        commands fail.
    :returns: True if total disk size exceeds 2TB. Returns False otherwise.
    """
    try:
        disksize_bytes, err = utils.execute('blockdev', '--getsize64',
                                            device,
                                            use_standard_locale=True,
                                            run_as_root=True)
    except (processutils.UnknownArgumentError,
            processutils.ProcessExecutionError, OSError) as e:
        msg = ('Failed to get size of disk %(disk)s for node %(node)s. '
               'Error: %(error)s' %
               {'disk': device, 'node': node_uuid, 'error': e})
        LOG.error(msg)
        raise exception.InstanceDeployFailure(msg)

    disksize_mb = int(disksize_bytes.strip()) // 1024 // 1024

    return disksize_mb > MAX_DISK_SIZE_MB_SUPPORTED_BY_MBR


def get_partition(device, uuid):
    """Find the partition of a given device."""
    LOG.debug("Find the partition %(uuid)s on device %(dev)s",
              {'dev': device, 'uuid': uuid})

    try:
        ipa_utils.rescan_device(device)
        lsblk = utils.execute(
            'lsblk', '-PbioKNAME,UUID,PARTUUID,TYPE,LABEL', device)
        report = lsblk[0]
        for line in report.split('\n'):
            part = {}
            # Split into KEY=VAL pairs
            vals = shlex.split(line)
            for key, val in (v.split('=', 1) for v in vals):
                part[key] = val.strip()
            # Ignore non partition
            if part.get('TYPE') not in ['md', 'part']:
                # NOTE(TheJulia): This technically creates an edge failure
                # case where a filesystem on a whole block device sans
                # partitioning would behave differently.
                continue

            if part.get('UUID') == uuid:
                LOG.debug("Partition %(uuid)s found on device "
                          "%(dev)s", {'uuid': uuid, 'dev': device})
                return '/dev/' + part.get('KNAME')
            if part.get('PARTUUID') == uuid:
                LOG.debug("Partition %(uuid)s found on device "
                          "%(dev)s", {'uuid': uuid, 'dev': device})
                return '/dev/' + part.get('KNAME')
            if part.get('LABEL') == uuid:
                LOG.debug("Partition %(uuid)s found on device "
                          "%(dev)s", {'uuid': uuid, 'dev': device})
                return '/dev/' + part.get('KNAME')
        else:
            # NOTE(TheJulia): We may want to consider moving towards using
            # findfs in the future, if we're comfortable with the execution
            # and interaction. There is value in either way though.
            # NOTE(rg): alternative: blkid -l -t UUID=/PARTUUID=
            try:
                findfs, stderr = utils.execute('findfs', 'UUID=%s' % uuid)
                return findfs.strip()
            except processutils.ProcessExecutionError as e:
                LOG.debug('First fallback detection attempt for locating '
                          'partition via UUID %(uuid)s failed. '
                          'Error: %(err)s',
                          {'uuid': uuid,
                           'err': e})
                try:
                    findfs, stderr = utils.execute(
                        'findfs', 'PARTUUID=%s' % uuid)
                    return findfs.strip()
                except processutils.ProcessExecutionError as e:
                    LOG.debug('Secondary fallback detection attempt for '
                              'locating partition via UUID %(uuid)s failed. '
                              'Error: %(err)s',
                              {'uuid': uuid,
                               'err': e})

            # Last fallback: In case we cannot find the partition by UUID
            # and the deploy device is an md device, we check if the md
            # device has a partition (which we assume to contain the root fs).
            if hardware.is_md_device(device):
                md_partition = device + 'p1'
                if (os.path.exists(md_partition)
                        and stat.S_ISBLK(os.stat(md_partition).st_mode)):
                    LOG.debug("Found md device with partition %s",
                              md_partition)
                    return md_partition
                else:
                    LOG.debug('Could not find partition %(part)s on md '
                              'device %(dev)s',
                              {'part': md_partition,
                               'dev': device})

            # Partition not found, time to escalate.
            error_msg = ("No partition with UUID %(uuid)s found on "
                         "device %(dev)s" % {'uuid': uuid, 'dev': device})
            LOG.error(error_msg)
            raise errors.DeviceNotFound(error_msg)
    except processutils.ProcessExecutionError as e:
        error_msg = ('Finding the partition with UUID %(uuid)s on '
                     'device %(dev)s failed with %(err)s' %
                     {'uuid': uuid, 'dev': device, 'err': e})
        LOG.error(error_msg)
        raise errors.CommandExecutionError(error_msg)



# Copyright 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

from oslo_log import log

from ironic_python_agent import errors
from ironic_python_agent import hardware

LOG = log.getLogger()


class ExampleBusinessLogicHardwareManager(hardware.HardwareManager):
    """Example hardware manager to enforce business logic"""

    # All hardware managers have a name and a version.
    # Version should be bumped anytime a change is introduced. This will
    # signal to Ironic that if automatic node cleaning is in progress to
    # restart it from the beginning, to ensure consistency. The value can
    # be anything; it's checked for equality against previously seen
    # name:manager pairs.
    HARDWARE_MANAGER_NAME = 'ExampleBusinessLogicHardwareManager'
    HARDWARE_MANAGER_VERSION = '1'

    def evaluate_hardware_support(self):
        """Declare level of hardware support provided.

        Since this example is explicitly about enforcing business logic during
        cleaning, we want to return a static value.

        :returns: HardwareSupport level for this manager.
        """
        return hardware.HardwareSupport.SERVICE_PROVIDER

    def get_clean_steps(self, node, ports):
        """Get a list of clean steps with priority.

        Define any clean steps added by this manager here. These will be mixed
        with other loaded managers that support this hardware, and ordered by
        priority. Higher priority steps run earlier.

        Note that out-of-band clean steps may also be provided by Ironic.
        These will follow the same priority ordering even though they are not
        executed by IPA.

        There is *no guarantee whatsoever* that steps defined here will be
        executed by this HardwareManager. When it comes time to run these
        steps, they'll be called using dispatch_to_managers() just like any
        other IPA HardwareManager method. This means if they are unique to
        your hardware, they should be uniquely named. For example,
        upgrade_firmware would be a bad step name. Whereas
        upgrade_foobar_device_firmware would be better.

        :param node: The node object as provided by Ironic.
        :param ports: Port objects as provided by Ironic.
        :returns: A list of cleaning steps, as a list of dicts.
        """
        # While obviously you could actively run code here, generally this
        # should just return a static value, as any initialization and
        # detection should've been done in evaluate_hardware_support().
        return [{
            'step': 'companyx_verify_device_lifecycle',
            'priority': 472,
            # If you need Ironic to coordinate a reboot after this step
            # runs, but before continuing cleaning, this should be true.
            'reboot_requested': False,
            # If it's safe for Ironic to abort cleaning while this step
            # runs, this should be true.
            'abortable': True
        }]

    # Other examples of interesting cleaning steps for this kind of hardware
    # manager would include verifying node.properties matches current state of
    # the node, checking smart stats to ensure the disk is not soon to fail,
    # or enforcing security policies.
    def companyx_verify_device_lifecycle(self, node, ports):
        """Verify node is not beyond useful life of 3 years."""
        create_date = node.get('created_at')
        if create_date is not None:
            server_age = time.time() - time.mktime(time.strptime(create_date))
            if server_age > (60 * 60 * 24 * 365 * 3):
                raise errors.CleaningError(
                    'Server is too old to pass cleaning!')
            else:
                LOG.info('Node is %s seconds old, younger than 3 years, '
                         'cleaning passes.', server_age)

    def get_deploy_steps(self, node, ports):
        """Get a list of deploy steps with priority.

        Returns a list of steps. Each step is represented by a dict::

          {
           'interface': the name of the driver interface that should execute
                        the step.
           'step': the HardwareManager function to call.
           'priority': the order steps will be run in. Ironic will sort all
                       the deploy steps from all the drivers, with the largest
                       priority step being run first. If priority is set to 0,
                       the step will not be run during deployment
                       automatically, but may be requested via deploy
                       templates.
           'reboot_requested': Whether the agent should request Ironic reboots
                               the node via the power driver after the
                               operation completes.
           'argsinfo': arguments specification.
          }


        Deploy steps are executed by similar logic to clean steps, but during
        deploy time. The following priority ranges should be used:

        * 81 to 99 - deploy steps to run before the image is written.
        * 61 to 79 - deploy steps to run after the image is written but before
          the bootloader is installed.
        * 41 to 59 - steps to run after the image is written and the bootloader
          is installed.

        If priority is zero, the step must be explicitly selected via
        an applied deploy template.

        Note that each deploy step makes deployments longer. Try to use clean
        steps for anything that is not required to be run just before an
        instance is ready.

        :param node: Ironic node object
        :param ports: list of Ironic port objects
        :return: a list of deploying steps, where each step is described as a
                 dict as defined above

        """
        return [
            {
                'step': 'companyx_verify_memory',
                'priority': 90,  # always run before the image is written
                'interface': 'deploy',
                'reboot_requested': False,
            },
            {
                'step': 'companyx_apply_something',
                'priority': 0,  # only run explicitly
                'interface': 'deploy',
                'reboot_requested': False,
                'argsinfo': {
                    "required_value": {
                        "description": "An example required argument.",
                        "required": True,
                    },
                    "optional_value": {
                        "description": "An example optional argument.",
                        "required": False,
                    }

                }
            },
        ]

    def companyx_verify_memory(self, node, ports):
        expected = node.get('properties', {}).get('memory_mb')
        if expected is None:
            LOG.warning('The node does not have memory, cannot verify')
            return
        else:
            expected = int(expected)

        # Use dispatch_to_managers to avoid tight coupling between hardware
        # managers. It would make sense even if this hardware manager
        # implemented get_memory, because a more specific hardware managers
        # could do it better.
        current = hardware.dispatch_to_managers('get_memory')
        if current.physical_mb < expected:
            # Raising an exception will fail deployment and set the node's
            # last_error accordingly.
            raise errors.DeploymentError(
                'Memory too low, expected {}, got {}'.format(
                    expected, current.physical_mb))

    # Make sure to provide default values for optional arguments.
    def companyx_apply_something(self, node, ports, required_value,
                                 optional_value=None):
        LOG.info('apply_something called with '
                 'required_value=%(required_value)s and '
                 'optional_value=%(optional_value)s',
                 {'required_value': required_value,
                  'optional_value': optional_value})



# Copyright (C) 2017 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_concurrency import processutils

from ironic_python_agent import dmi_inspector
from ironic_python_agent.tests.unit import base
from ironic_python_agent.tests.unit import dmi_inspector_data as dmi_data
from ironic_python_agent import utils


class TestCollectDmidecodeInfo(base.IronicAgentTest):
    def setUp(self):
        super(TestCollectDmidecodeInfo, self).setUp()
        self.data = {}
        self.failures = utils.AccumulatedFailures()

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_dmidecode_info_ok(self, mock_execute):
        mock_execute.return_value = dmi_data.DMI_DATA

        dmi_inspector.collect_dmidecode_info(self.data, None)

        for key in ('bios', 'memory', 'cpu'):
            self.assertTrue(self.data['dmi'][key])

        self.assertEqual(dmi_data.DMI_OUTPUT, self.data)

        mock_execute.assert_called_once_with('dmidecode', '-t', 'bios',
                                             '-t', 'processor', '-t', 'memory')

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_dmidecode_info_bad_data(self, mock_execute):
        mock_execute.return_value = ("""Handle 0x0000\nFoo\nBar: Baz\n""", "")
        expected = {'dmi': {'bios': {}, 'cpu': [], 'memory': {'devices': []}}}

        dmi_inspector.collect_dmidecode_info(self.data, None)

        self.assertEqual(expected, self.data)

        mock_execute.assert_called_once_with('dmidecode', '-t', 'bios',
                                             '-t', 'processor', '-t', 'memory')

    @mock.patch.object(utils, 'execute', autospec=True)
    def test_dmidecode_info_failure(self, mock_execute):
        mock_execute.side_effect = processutils.ProcessExecutionError()

        dmi_inspector.collect_dmidecode_info(self.data, self.failures)

        self.assertTrue(self.failures)
        self.assertNotIn('dmi', self.data)
        mock_execute.assert_called_once_with('dmidecode', '-t', 'bios',
                                             '-t', 'processor', '-t', 'memory')

    def test_parse_dmi_bios(self):
        inputdata = dmi_data.BIOS_DATA[0]
        expected = dmi_data.BIOS_OUTPUT

        ret = dmi_inspector.parse_dmi(inputdata)
        self.assertEqual(expected, ret)

    def test_parse_dmi_cpu(self):
        inputdata = dmi_data.CPU_DATA[0]
        expected = dmi_data.CPU_OUTPUT

        ret = dmi_inspector.parse_dmi(inputdata)
        self.assertEqual(expected, ret)

    def test_parse_dmi_memory(self):
        inputdata = dmi_data.MEMORY_DATA[0]
        expected = dmi_data.MEMORY_OUTPUT

        ret = dmi_inspector.parse_dmi(inputdata)
        self.assertEqual(expected, ret)

    def test_save_data(self):
        dmi_info = {}
        dmi_info['bios'] = {}
        dmi_info['cpu'] = []
        dmi_info['memory'] = {}
        dmi_info['memory']['devices'] = {}
        mem = [{'Handle': 'handle', 'Number Of Devices': '2'},
               {'Handle': 'handle', 'Number Of Devices': '2'}]
        devices = [{'bar': 'foo'}, {'bar': 'foo'}]
        expected = {'bios': {},
                    'cpu': [],
                    'memory': {'Number Of Devices': 4,
                               'devices': [{'bar': 'foo'}, {'bar': 'foo'}]}}

        ret = dmi_inspector._save_data(dmi_info, mem, devices)
        self.assertEqual(expected, ret)

    def test_save_data_error_number_of_devices(self):
        dmi_info = {}
        dmi_info['bios'] = {}
        dmi_info['cpu'] = []
        dmi_info['memory'] = {}
        dmi_info['memory']['devices'] = {}

        self.assertRaises(KeyError,
                          dmi_inspector._save_data,
                          dmi_info,
                          [{'foo': 'bar', 'Handle': '0x10'}],
                          [{'bar': 'foo'}, {'bar': 'foo'}])

    def test_save_data_error_handle(self):
        dmi_info = {}
        dmi_info['bios'] = {}
        dmi_info['cpu'] = []
        dmi_info['memory'] = {}
        dmi_info['memory']['devices'] = {}

        self.assertRaises(KeyError,
                          dmi_inspector._save_data,
                          dmi_info,
                          [{'foo': 'bar', 'Number Of Devices': '2'}],
                          [{'bar': 'foo'}, {'bar': 'foo'}])



# Copyright 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid

from ironic_lib import exception as lib_exc


class Serializable(object):
    """Base class for things that can be serialized."""
    serializable_fields = ()

    def serialize(self):
        """Turn this object into a dict."""
        return dict((f, getattr(self, f)) for f in self.serializable_fields)


class SerializableComparable(Serializable):
    """A Serializable class which supports some comparison operators

    This class supports the '__eq__' and '__ne__' comparison operators, but
    intentionally disables the '__hash__' operator as some child classes may be
    mutable.  The addition of these comparison operators is mainly used to
    assist with unit testing.
    """

    __hash__ = None

    def __eq__(self, other):
        return self.serialize() == other.serialize()

    def __ne__(self, other):
        return self.serialize() != other.serialize()


def serialize_lib_exc(exc):
    """Serialize an ironic-lib exception."""
    return {'type': exc.__class__.__name__,
            'code': exc.code,
            'message': str(exc),
            'details': ''}


class RESTJSONEncoder(json.JSONEncoder):
    """A slightly customized JSON encoder."""
    def encode(self, o):
        """Turn an object into JSON.

        Appends a newline to responses when configured to pretty-print,
        in order to make use of curl less painful from most shells.
        """
        delimiter = ''

        # if indent is None, newlines are still inserted, so we should too.
        if self.indent is not None:
            delimiter = '\n'

        return super(RESTJSONEncoder, self).encode(o) + delimiter

    def default(self, o):
        """Turn an object into a serializable object.

        In particular, by calling :meth:`.Serializable.serialize` on `o`.
        """
        if isinstance(o, Serializable):
            return o.serialize()
        elif isinstance(o, uuid.UUID):
            return str(o)
        elif isinstance(o, lib_exc.IronicException):
            return serialize_lib_exc(o)
        else:
            return json.JSONEncoder.default(self, o)



# Copyright 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import ipaddress
import random
import socket
import threading
import time
from urllib import parse as urlparse

import eventlet
from ironic_lib import exception as lib_exc
from ironic_lib import mdns
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log
import pkg_resources

from ironic_python_agent.api import app
from ironic_python_agent import config
from ironic_python_agent import encoding
from ironic_python_agent import errors
from ironic_python_agent.extensions import base
from ironic_python_agent import hardware
from ironic_python_agent import inspector
from ironic_python_agent import ironic_api_client
from ironic_python_agent import utils

LOG = log.getLogger(__name__)

# Time(in seconds) to wait for any of the interfaces to be up
# before lookup of the node is attempted
NETWORK_WAIT_TIMEOUT = 60

# Time(in seconds) to wait before reattempt
NETWORK_WAIT_RETRY = 5

cfg.CONF.import_group('metrics', 'ironic_lib.metrics_utils')
cfg.CONF.import_group('metrics_statsd', 'ironic_lib.metrics_statsd')

Host = collections.namedtuple('Host', ['hostname', 'port'])


def _time():
    """Wraps time.time() for simpler testing."""
    return time.time()


class IronicPythonAgentStatus(encoding.Serializable):
    """Represents the status of an agent."""

    serializable_fields = ('started_at', 'version')

    def __init__(self, started_at, version):
        self.started_at = started_at
        self.version = version


class IronicPythonAgentHeartbeater(threading.Thread):
    """Thread that periodically heartbeats to Ironic."""

    # If we could wait at most N seconds between heartbeats (or in case of an
    # error) we will instead wait r x N seconds, where r is a random value
    # between these multipliers.
    min_jitter_multiplier = 0.3
    max_jitter_multiplier = 0.6

    def __init__(self, agent):
        """Initialize the heartbeat thread.

        :param agent: an :class:`ironic_python_agent.agent.IronicPythonAgent`
                      instance.
        """
        super(IronicPythonAgentHeartbeater, self).__init__()
        self.agent = agent
        self.stop_event = threading.Event()
        self.api = agent.api_client
        self.interval = 0
        self.heartbeat_forced = False
        self.previous_heartbeat = 0

    def run(self):
        """Start the heartbeat thread."""
        # The first heartbeat happens immediately
        LOG.info('Starting heartbeater')
        self.agent.set_agent_advertise_addr()

        while not self.stop_event.wait(min(self.interval, 5)):
            if self._heartbeat_expected():
                self.do_heartbeat()
            eventlet.sleep(0)

    def _heartbeat_expected(self):
        # Normal heartbeating
        if _time() > self.previous_heartbeat + self.interval:
            return True

        # Forced heartbeating, but once in 5 seconds
        if (self.heartbeat_forced
                and _time() > self.previous_heartbeat + 5):
            return True

    def do_heartbeat(self):
        """Send a heartbeat to Ironic."""
        try:
            self.api.heartbeat(
                uuid=self.agent.get_node_uuid(),
                advertise_address=self.agent.advertise_address,
                advertise_protocol=self.agent.advertise_protocol,
                generated_cert=self.agent.generated_cert,
            )
            LOG.info('heartbeat successful')
            self.heartbeat_forced = False
            self.previous_heartbeat = _time()
        except errors.HeartbeatConflictError:
            LOG.warning('conflict error sending heartbeat to %s',
                        self.agent.api_url)
        except Exception:
            LOG.exception('error sending heartbeat to %s', self.agent.api_url)
        finally:
            interval_multiplier = random.uniform(self.min_jitter_multiplier,
                                                 self.max_jitter_multiplier)
            self.interval = self.agent.heartbeat_timeout * interval_multiplier
            LOG.info('sleeping before next heartbeat, interval: %s',
                     self.interval)

    def force_heartbeat(self):
        self.heartbeat_forced = True

    def stop(self):
        """Stop the heartbeat thread."""
        LOG.info('stopping heartbeater')
        self.stop_event.set()
        return self.join()


class IronicPythonAgent(base.ExecuteCommandMixin):
    """Class for base agent functionality."""

    def __init__(self, api_url, advertise_address, listen_address,
                 ip_lookup_attempts, ip_lookup_sleep, network_interface,
                 lookup_timeout, lookup_interval, standalone, agent_token,
                 hardware_initialization_delay=0, advertise_protocol='http'):
        super(IronicPythonAgent, self).__init__()
        if bool(cfg.CONF.keyfile) != bool(cfg.CONF.certfile):
            LOG.warning("Only one of 'keyfile' and 'certfile' options is "
                        "defined in config file. Its value will be ignored.")
        self.ext_mgr = base.init_ext_manager(self)
        self.api_url = api_url
        if (not self.api_url or self.api_url == 'mdns') and not standalone:
            try:
                self.api_url, params = mdns.get_endpoint('baremetal')
            except lib_exc.ServiceLookupFailure:
                if self.api_url:
                    # mDNS explicitly requested, report failure.
                    raise
                else:
                    # implicit fallback to mDNS, do not fail (maybe we're only
                    # running inspection).
                    LOG.warning('Could not get baremetal endpoint from mDNS, '
                                'will not heartbeat')
            else:
                config.override(params)

        if self.api_url:
            self.api_client = ironic_api_client.APIClient(self.api_url)
            self.heartbeater = IronicPythonAgentHeartbeater(self)
        self.listen_address = listen_address
        self.advertise_address = advertise_address
        self.advertise_protocol = advertise_protocol
        self.version = pkg_resources.get_distribution('ironic-python-agent')\
            .version
        self.api = app.Application(self, cfg.CONF)
        self.heartbeat_timeout = None
        self.started_at = None
        self.node = None
        # lookup timeout in seconds
        self.lookup_timeout = lookup_timeout
        self.lookup_interval = lookup_interval
        self.ip_lookup_attempts = ip_lookup_attempts
        self.ip_lookup_sleep = ip_lookup_sleep
        self.network_interface = network_interface
        self.standalone = standalone
        self.hardware_initialization_delay = hardware_initialization_delay
        # IPA will stop serving requests and exit after this is set to False
        self.serve_api = True
        self.agent_token = agent_token
        # Allows this to be turned on by the conductor while running,
        # in the event of long running ramdisks where the conductor
        # got upgraded somewhere along the way.
        self.agent_token_required = cfg.CONF.agent_token_required
        self.generated_cert = None

    def get_status(self):
        """Retrieve a serializable status.

        :returns: a :class:`ironic_python_agent.agent.IronicPythonAgent`
                  instance describing the agent's status.
        """
        return IronicPythonAgentStatus(
            started_at=self.started_at,
            version=self.version
        )

    def validate_agent_token(self, token):
        # We did not get a token, i.e. None and
        # we've previously seen a token, which is
        # a mid-cluster upgrade case with long-running ramdisks.
        if (not token and self.agent_token
                and not self.agent_token_required):
            # TODO(TheJulia): Rip this out during or after the V cycle.
            LOG.warning('Agent token for requests are not required '
                        'by the conductor, yet we received a token. '
                        'Cluster may be mid-upgrade. Support to '
                        'not fail in this condition will be removed in '
                        'the Victoria development cycle.')
            # Tell the API everything is okay.
            return True

        return self.agent_token == token

    def _get_route_source(self, dest):
        """Get the IP address to send packages to destination."""
        try:
            out, _err = utils.execute('ip', 'route', 'get', dest)
        except (EnvironmentError, processutils.ProcessExecutionError) as e:
            LOG.warning('Cannot get route to host %(dest)s: %(err)s',
                        {'dest': dest, 'err': e})
            return

        try:
            source = out.strip().split('\n')[0].split('src')[1].split()[0]
        except IndexError:
            LOG.warning('No route to host %(dest)s, route record: %(rec)s',
                        {'dest': dest, 'rec': out})
            return

        try:
            if ipaddress.ip_address(source).is_link_local:
                LOG.info('Ignoring link-local source to %(dest)s: %(rec)s',
                         {'dest': dest, 'rec': out})
                return
        except ValueError as exc:
            LOG.warning('Invalid IP address %(addr)s returned as a route '
                        'to host %(dest)s: %(err)s',
                        {'dest': dest, 'addr': source, 'err': exc})

        return source

    def set_agent_advertise_addr(self):
        """Set advertised IP address for the agent, if not already set.

        If agent's advertised IP address is still default (None), try to
        find a better one.  If the agent's network interface is None, replace
        that as well.

        :raises: LookupAgentIPError if an IP address could not be found
        """
        if self.advertise_address.hostname is not None:
            return

        found_ip = None
        if self.network_interface is not None:
            # TODO(dtantsur): deprecate this
            found_ip = hardware.dispatch_to_managers('get_ipv4_addr',
                                                     self.network_interface)
        else:
            url = urlparse.urlparse(self.api_url)
            ironic_host = url.hostname
            # Try resolving it in case it's not an IP address
            try:
                ironic_host = socket.gethostbyname(ironic_host)
            except socket.gaierror:
                LOG.debug('Count not resolve %s, maybe no DNS', ironic_host)

            for attempt in range(self.ip_lookup_attempts):
                found_ip = self._get_route_source(ironic_host)
                if found_ip:
                    break

                time.sleep(self.ip_lookup_sleep)

        if found_ip:
            self.advertise_address = Host(hostname=found_ip,
                                          port=self.advertise_address.port)
        else:
            raise errors.LookupAgentIPError('Agent could not find a valid IP '
                                            'address.')

    def get_node_uuid(self):
        """Get UUID for Ironic node.

        If the agent has not yet heartbeated to Ironic, it will not have
        the UUID and this will raise an exception.

        :returns: A string containing the UUID for the Ironic node.
        :raises: UnknownNodeError if UUID is unknown.
        """
        if self.node is None or 'uuid' not in self.node:
            raise errors.UnknownNodeError()
        return self.node['uuid']

    def list_command_results(self):
        """Get a list of command results.

        :returns: list of :class:`ironic_python_agent.extensions.base.
                  BaseCommandResult` objects.
        """
        return list(self.command_results.values())

    def get_command_result(self, result_id):
        """Get a specific command result by ID.

        :returns: a :class:`ironic_python_agent.extensions.base.
                  BaseCommandResult` object.
        :raises: RequestedObjectNotFoundError if command with the given ID
                 is not found.
        """
        try:
            return self.command_results[result_id]
        except KeyError:
            raise errors.RequestedObjectNotFoundError('Command Result',
                                                      result_id)

    def force_heartbeat(self):
        if not self.standalone:
            self.heartbeater.force_heartbeat()

    def _wait_for_interface(self):
        """Wait until at least one interface is up."""

        wait_till = time.time() + NETWORK_WAIT_TIMEOUT
        while time.time() < wait_till:
            interfaces = hardware.dispatch_to_managers(
                'list_network_interfaces')
            if not any(ifc.mac_address for ifc in interfaces):
                LOG.debug('Network is not up yet. '
                          'No valid interfaces found, retrying ...')
                time.sleep(NETWORK_WAIT_RETRY)
            else:
                break

        else:
            LOG.warning("No valid network interfaces found. "
                        "Node lookup will probably fail.")

    def _start_auto_tls(self):
        # NOTE(dtantsur): if listen_tls is True, assume static TLS
        # configuration and don't auto-generate anything.
        if cfg.CONF.listen_tls or not cfg.CONF.enable_auto_tls:
            LOG.debug('Automated TLS is disabled')
            return None, None

        if not self.api_url or not self.api_client.supports_auto_tls():
            LOG.warning('Ironic does not support automated TLS')
            return None, None

        self.set_agent_advertise_addr()

        LOG.info('Generating TLS parameters automatically for IP %s',
                 self.advertise_address.hostname)
        tls_info = hardware.dispatch_to_managers(
            'generate_tls_certificate', self.advertise_address.hostname)
        self.generated_cert = tls_info.text
        self.advertise_protocol = 'https'
        return tls_info.path, tls_info.private_key_path

    def serve_ipa_api(self):
        """Serve the API until an extension terminates it."""
        cert_file, key_file = self._start_auto_tls()
        self.api.start(cert_file, key_file)
        if not self.standalone and self.api_url:
            # Don't start heartbeating until the server is listening
            self.heartbeater.start()
        try:
            while self.serve_api:
                eventlet.sleep(0.1)
        except KeyboardInterrupt:
            LOG.info('Caught keyboard interrupt, exiting')
        self.api.stop()

    def process_lookup_data(self, content):
        """Update agent configuration from lookup data."""

        self.node = content['node']
        LOG.info('Lookup succeeded, node UUID is %s',
                 self.node['uuid'])
        hardware.cache_node(self.node)
        self.heartbeat_timeout = content['config']['heartbeat_timeout']

        # Update config with values from Ironic
        config = content.get('config', {})
        if config.get('metrics'):
            for opt, val in config.items():
                setattr(cfg.CONF.metrics, opt, val)
        if config.get('metrics_statsd'):
            for opt, val in config.items():
                setattr(cfg.CONF.metrics_statsd, opt, val)
        if config.get('agent_token_required'):
            self.agent_token_required = True
        token = config.get('agent_token')
        if token:
            if len(token) >= 32:
                LOG.debug('Agent token recorded as designated by '
                          'the ironic installation.')
                self.agent_token = token
                # set with-in the API client.
                if not self.standalone:
                    self.api_client.agent_token = token
            elif token == '******':
                LOG.warning('The agent token has already been '
                            'retrieved. IPA may not operate as '
                            'intended and the deployment may fail '
                            'depending on settings in the ironic '
                            'deployment.')
                if not self.agent_token and self.agent_token_required:
                    LOG.error('Ironic is signaling that agent tokens '
                              'are required, however we do not have '
                              'a token on file. '
                              'This is likely **FATAL**.')
            else:
                LOG.info('An invalid token was received.')
        if self.agent_token and not self.standalone:
            # Explicitly set the token in our API client before
            # starting heartbeat operations.
            self.api_client.agent_token = self.agent_token

    def run(self):
        """Run the Ironic Python Agent."""
        LOG.info('Starting ironic-python-agent version: %s',
                 self.version)
        # Get the UUID so we can heartbeat to Ironic. Raises LookupNodeError
        # if there is an issue (uncaught, restart agent)
        self.started_at = _time()
        # Attempt to sync the software clock
        utils.sync_clock(ignore_errors=True)

        # Cached hw managers at runtime, not load time. See bug 1490008.
        hardware.get_managers()
        # Operator-settable delay before hardware actually comes up.
        # Helps with slow RAID drivers - see bug 1582797.
        if self.hardware_initialization_delay > 0:
            LOG.info('Waiting %d seconds before proceeding',
                     self.hardware_initialization_delay)
            time.sleep(self.hardware_initialization_delay)

        if not self.standalone:
            # Inspection should be started before call to lookup, otherwise
            # lookup will fail due to unknown MAC.
            uuid = None
            # We can't try to inspect or heartbeat until we have valid
            # interfaces to perform those actions over.
            self._wait_for_interface()

            if cfg.CONF.inspection_callback_url:
                try:
                    # Attempt inspection. This may fail, and previously
                    # an error would be logged.
                    uuid = inspector.inspect()
                except errors.InspectionError as e:
                    LOG.error('Failed to perform inspection: %s', e)

            if self.api_url:
                content = self.api_client.lookup_node(
                    hardware_info=hardware.list_hardware_info(use_cache=True),
                    timeout=self.lookup_timeout,
                    starting_interval=self.lookup_interval,
                    node_uuid=uuid)
                LOG.debug('Received lookup results: %s', content)
                self.process_lookup_data(content)
                # Save the API url in case we need it later.
                hardware.save_api_client(
                    self.api_client, self.lookup_timeout,
                    self.lookup_interval)

            elif cfg.CONF.inspection_callback_url:
                LOG.info('No ipa-api-url configured, Heartbeat and lookup '
                         'skipped for inspector.')
            else:
                # NOTE(TheJulia): Once communication flow capability is
                # able to be driven solely from the conductor, this is no
                # longer a major issue.
                LOG.error('Neither ipa-api-url nor inspection_callback_url'
                          'found, please check your pxe append parameters.')

        self.serve_ipa_api()

        if not self.standalone and self.api_url:
            self.heartbeater.stop()



# Copyright 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

from ironic_lib import metrics_utils
from oslo_log import log
from oslo_service import wsgi
import werkzeug
from werkzeug import exceptions as http_exc
from werkzeug import routing
from werkzeug.wrappers import json as http_json

from ironic_python_agent import encoding


LOG = log.getLogger(__name__)
_CUSTOM_MEDIA_TYPE = 'application/vnd.openstack.ironic-python-agent.v1+json'
_DOCS_URL = 'https://docs.openstack.org/ironic-python-agent'


class Request(werkzeug.Request, http_json.JSONMixin):
    """Custom request class with JSON support."""


def jsonify(value, status=200):
    """Convert value to a JSON response using the custom encoder."""
    encoder = encoding.RESTJSONEncoder()
    data = encoder.encode(value)
    return werkzeug.Response(data, status=status, mimetype='application/json')


def make_link(url, rel_name, resource='', resource_args='',
              bookmark=False, type_=None):
    if rel_name == 'describedby':
        url = _DOCS_URL
        type_ = 'text/html'
    elif rel_name == 'bookmark':
        bookmark = True

    template = ('%(root)s/%(resource)s' if bookmark
                else '%(root)s/v1/%(resource)s')
    template += ('%(args)s'
                 if resource_args.startswith('?') or not resource_args
                 else '/%(args)s')

    result = {'href': template % {'root': url,
                                  'resource': resource,
                                  'args': resource_args},
              'rel': rel_name}
    if type_:
        result['type'] = type_
    return result


def version(url):
    return {
        'id': 'v1',
        'links': [
            make_link(url, 'self', 'v1', bookmark=True),
            make_link(url, 'describedby', bookmark=True),
        ],
    }


# Emulate WSME format
def format_exception(value):
    code = getattr(value, 'status_code', None) or getattr(value, 'code', 500)
    return {
        'faultcode': 'Server' if code >= 500 else 'Client',
        'faultstring': str(value),
    }


class Application(object):

    def __init__(self, agent, conf):
        """Set up the API app.

        :param agent: an :class:`ironic_python_agent.agent.IronicPythonAgent`
                      instance.
        :param conf: configuration object.
        """
        self.agent = agent
        self.service = None
        self._conf = conf
        self.url_map = routing.Map([
            routing.Rule('/', endpoint='root', methods=['GET']),
            routing.Rule('/v1/', endpoint='v1', methods=['GET']),
            routing.Rule('/v1/status', endpoint='status', methods=['GET']),
            routing.Rule('/v1/commands/', endpoint='list_commands',
                         methods=['GET']),
            routing.Rule('/v1/commands/<cmd>', endpoint='get_command',
                         methods=['GET']),
            routing.Rule('/v1/commands/', endpoint='run_command',
                         methods=['POST']),
            # Use the default version (i.e. v1) when the version is missing
            routing.Rule('/status', endpoint='status', methods=['GET']),
            routing.Rule('/commands/', endpoint='list_commands',
                         methods=['GET']),
            routing.Rule('/commands/<cmd>', endpoint='get_command',
                         methods=['GET']),
            routing.Rule('/commands/', endpoint='run_command',
                         methods=['POST']),
        ])

    def __call__(self, environ, start_response):
        """WSGI entry point."""
        try:
            request = Request(environ)
            adapter = self.url_map.bind_to_environ(request.environ)
            endpoint, values = adapter.match()
            response = getattr(self, "api_" + endpoint)(request, **values)
        except Exception as exc:
            response = self.handle_exception(environ, exc)
        return response(environ, start_response)

    def start(self, tls_cert_file=None, tls_key_file=None):
        """Start the API service in the background."""
        if tls_cert_file and tls_key_file:
            self._conf.set_override('cert_file', tls_cert_file, group='ssl')
            self._conf.set_override('key_file', tls_key_file, group='ssl')
            use_tls = True
        else:
            use_tls = self._conf.listen_tls

        self.service = wsgi.Server(self._conf, 'ironic-python-agent', app=self,
                                   host=self.agent.listen_address.hostname,
                                   port=self.agent.listen_address.port,
                                   use_ssl=use_tls)
        self.service.start()
        LOG.info('Started API service on port %s',
                 self.agent.listen_address.port)

    def stop(self):
        """Stop the API service."""
        LOG.debug("Stopping the API service.")
        if self.service is None:
            return
        self.service.stop()
        self.service = None
        LOG.info('Stopped API service on port %s',
                 self.agent.listen_address.port)

    def handle_exception(self, environ, exc):
        """Handle an exception during request processing."""
        if isinstance(exc, http_exc.HTTPException):
            if exc.code and exc.code < 400:
                return exc  # redirect
            resp = exc.get_response(environ)
            resp.data = json.dumps(format_exception(exc))
            resp.content_type = 'application/json'
            return resp
        else:
            formatted = format_exception(exc)
            if formatted['faultcode'] == 'Server':
                LOG.exception('Internal server error: %s', exc)
            return jsonify(formatted, status=getattr(exc, 'status_code', 500))

    def api_root(self, request):
        url = request.url_root.rstrip('/')
        return jsonify({
            'name': 'OpenStack Ironic Python Agent API',
            'description': ('Ironic Python Agent is a provisioning agent for '
                            'OpenStack Ironic'),
            'versions': [version(url)],
            'default_version': version(url),
        })

    def api_v1(self, request):
        url = request.url_root.rstrip('/')
        return jsonify(dict({
            'commands': [
                make_link(url, 'self', 'commands'),
                make_link(url, 'bookmark', 'commands'),
            ],
            'status': [
                make_link(url, 'self', 'status'),
                make_link(url, 'bookmark', 'status'),
            ],
            'media_types': [
                {'base': 'application/json',
                 'type': _CUSTOM_MEDIA_TYPE},
            ],
        }, **version(url)))

    def api_status(self, request):
        with metrics_utils.get_metrics_logger(__name__).timer('get_status'):
            status = self.agent.get_status()
            return jsonify(status)

    def api_list_commands(self, request):
        with metrics_utils.get_metrics_logger(__name__).timer('list_commands'):
            results = self.agent.list_command_results()
            return jsonify({'commands': results})

    def api_get_command(self, request, cmd):
        with metrics_utils.get_metrics_logger(__name__).timer('get_command'):
            result = self.agent.get_command_result(cmd)
            wait = request.args.get('wait')

            if wait and wait.lower() == 'true':
                result.join()

            return jsonify(result)

    def api_run_command(self, request):
        body = request.get_json(force=True)
        if ('name' not in body or 'params' not in body
                or not isinstance(body['params'], dict)):
            raise http_exc.BadRequest('Missing or invalid name or params')

        token = request.args.get('agent_token', None)
        if not self.agent.validate_agent_token(token):
            raise http_exc.Unauthorized(
                'Token invalid.')
        with metrics_utils.get_metrics_logger(__name__).timer('run_command'):
            result = self.agent.execute_command(body['name'], **body['params'])
            wait = request.args.get('wait')
            if wait and wait.lower() == 'true':
                result.join()
            return jsonify(result)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import stat
import tempfile
from unittest import mock

from ironic_python_agent import errors
from ironic_python_agent import inject_files
from ironic_python_agent.tests.unit import base


@mock.patch('ironic_lib.utils.mounted', autospec=True)
@mock.patch('ironic_lib.disk_utils.list_partitions', autospec=True)
@mock.patch('ironic_python_agent.hardware.dispatch_to_managers',
            lambda _call: '/dev/fake')
class TestFindPartitionWithPath(base.IronicAgentTest):

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.tempdir))

    def test_found(self, mock_list_parts, mock_mount):
        mock_list_parts.return_value = [
            {'number': 1, 'flags': 'lvm'},
            {'number': 2, 'flags': 'boot'},
        ]
        mock_mount.return_value.__enter__.return_value = self.tempdir
        expected = os.path.join(self.tempdir, "some/path")
        os.makedirs(expected)

        with inject_files.find_partition_with_path("/some/path") as path:
            self.assertEqual(expected, path)

        mock_mount.assert_called_once_with('/dev/fake2')

    def test_found_with_dev(self, mock_list_parts, mock_mount):
        mock_list_parts.return_value = [
            {'number': 1, 'flags': 'lvm'},
            {'number': 2, 'flags': 'boot'},
        ]
        mock_mount.return_value.__enter__.return_value = self.tempdir
        expected = os.path.join(self.tempdir, "some/path")
        os.makedirs(expected)

        with inject_files.find_partition_with_path("/some/path",
                                                   "/dev/nvme0n1") as path:
            self.assertEqual(expected, path)

        mock_mount.assert_called_once_with('/dev/nvme0n1p2')

    def test_not_found(self, mock_list_parts, mock_mount):
        mock_list_parts.return_value = [
            {'number': 1, 'flags': 'lvm'},
            {'number': 2, 'flags': 'boot'},
            {'number': 3, 'flags': ''},
        ]
        mock_mount.return_value.__enter__.return_value = self.tempdir

        self.assertRaises(
            errors.DeviceNotFound,
            inject_files.find_partition_with_path("/some/path").__enter__)

        mock_mount.assert_has_calls([
            mock.call('/dev/fake2'),
            mock.call('/dev/fake3'),
        ], any_order=True)


class TestFindAndMountPath(base.IronicAgentTest):

    @mock.patch.object(inject_files, 'find_partition_with_path', autospec=True)
    def test_without_on(self, mock_find_part):
        mock_find_part.return_value.__enter__.return_value = '/mount/path'
        with inject_files._find_and_mount_path('/etc/sysctl.d/my.conf',
                                               None, '/dev/fake') as result:
            # "etc" is included in a real result of find_partition_with_path
            self.assertEqual('/mount/path/sysctl.d/my.conf', result)
        mock_find_part.assert_called_once_with('etc', '/dev/fake')

    def test_without_on_wrong_path(self):
        self.assertRaises(
            errors.InvalidCommandParamsError,
            inject_files._find_and_mount_path('/etc', None,
                                              '/dev/fake').__enter__)

    @mock.patch('ironic_lib.utils.mounted', autospec=True)
    def test_with_on_as_path(self, mock_mount):
        mock_mount.return_value.__enter__.return_value = '/mount/path'
        with inject_files._find_and_mount_path('/etc/sysctl.d/my.conf',
                                               '/dev/on',
                                               '/dev/fake') as result:
            self.assertEqual('/mount/path/etc/sysctl.d/my.conf', result)
        mock_mount.assert_called_once_with('/dev/on')

    @mock.patch('ironic_lib.utils.mounted', autospec=True)
    def test_with_on_as_number(self, mock_mount):
        mock_mount.return_value.__enter__.return_value = '/mount/path'
        with inject_files._find_and_mount_path('/etc/sysctl.d/my.conf',
                                               2, '/dev/fake') as result:
            self.assertEqual('/mount/path/etc/sysctl.d/my.conf', result)
        mock_mount.assert_called_once_with('/dev/fake2')

    @mock.patch('ironic_lib.utils.mounted', autospec=True)
    def test_with_on_as_number_nvme(self, mock_mount):
        mock_mount.return_value.__enter__.return_value = '/mount/path'
        with inject_files._find_and_mount_path('/etc/sysctl.d/my.conf',
                                               2, '/dev/nvme0n1') as result:
            self.assertEqual('/mount/path/etc/sysctl.d/my.conf', result)
        mock_mount.assert_called_once_with('/dev/nvme0n1p2')


@mock.patch.object(inject_files, '_find_and_mount_path', autospec=True)
class TestInjectOne(base.IronicAgentTest):

    def setUp(self):
        super().setUp()
        self.tempdir = tempfile.mkdtemp()
        self.addCleanup(lambda: shutil.rmtree(self.tempdir))
        self.dirpath = os.path.join(self.tempdir, 'dir1', 'dir2')
        self.path = os.path.join(self.dirpath, 'file.name')

        self.http_get = mock.MagicMock()
        self.http_get.return_value.__enter__.return_value = iter(
            [b'con', b'tent', b''])

        self.node = {'uuid': '1234'}
        self.ports = [{'address': 'aabb'}]

    def test_delete(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'deleted': True}
        os.makedirs(self.dirpath)
        with open(self.path, 'wb') as fp:
            fp.write(b'content')

        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        self.assertFalse(os.path.exists(self.path))
        self.assertTrue(os.path.isdir(self.dirpath))
        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_not_called()

    def test_delete_not_exists(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'deleted': True}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        self.assertFalse(os.path.exists(self.path))
        self.assertFalse(os.path.isdir(self.dirpath))
        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_not_called()

    def test_plain_content(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA=='}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())
        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_not_called()

    def test_plain_content_with_on(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA==',
              'partition': '/dev/sda1'}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())
        mock_find_and_mount.assert_called_once_with(fl['path'], '/dev/sda1',
                                                    '/dev/root')
        self.http_get.assert_not_called()

    def test_plain_content_with_modes(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA==',
              'mode': 0o602, 'dirmode': 0o703}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())
        self.assertEqual(0o602, stat.S_IMODE(os.stat(self.path).st_mode))
        self.assertEqual(0o703, stat.S_IMODE(os.stat(self.dirpath).st_mode))

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_not_called()

    def test_plain_content_with_modes_exists(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA==',
              'mode': 0o602, 'dirmode': 0o703}
        os.makedirs(self.dirpath)
        with open(self.path, 'wb') as fp:
            fp.write(b"I'm not a cat, I'm a lawyer")

        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())
        self.assertEqual(0o602, stat.S_IMODE(os.stat(self.path).st_mode))
        # Exising directories do not change their permissions
        self.assertNotEqual(0o703, stat.S_IMODE(os.stat(self.dirpath).st_mode))

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_not_called()

    @mock.patch.object(os, 'chown', autospec=True)
    def test_plain_content_with_owner(self, mock_chown, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA==',
              'owner': 42}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        mock_chown.assert_called_once_with(self.path, 42, -1)
        self.http_get.assert_not_called()

    @mock.patch.object(os, 'chown', autospec=True)
    def test_plain_content_with_owner_and_group(self, mock_chown,
                                                mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name', 'content': 'Y29udGVudA==',
              'owner': 0, 'group': 0}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        mock_chown.assert_called_once_with(self.path, 0, 0)
        self.http_get.assert_not_called()

    def test_url(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name',
              'content': 'http://example.com/path'}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_called_once_with('http://example.com/path')

    def test_url_formatting(self, mock_find_and_mount):
        mock_find_and_mount.return_value.__enter__.return_value = self.path

        fl = {'path': '/etc/dir1/dir2/file.name',
              'content': 'http://example.com/{node[uuid]}/{ports[0][address]}'}
        inject_files._inject_one(self.node, self.ports, fl,
                                 '/dev/root', self.http_get)

        with open(self.path, 'rb') as fp:
            self.assertEqual(b'content', fp.read())

        mock_find_and_mount.assert_called_once_with(fl['path'], None,
                                                    '/dev/root')
        self.http_get.assert_called_once_with('http://example.com/1234/aabb')


@mock.patch('ironic_python_agent.hardware.dispatch_to_managers',
            lambda _call: '/dev/root')
@mock.patch.object(inject_files, '_inject_one', autospec=True)
class TestInjectFiles(base.IronicAgentTest):

    def test_empty(self, mock_inject):
        node = {
            'properties': {}
        }

        inject_files.inject_files(node, [mock.sentinel.port], [])
        mock_inject.assert_not_called()

    def test_ok(self, mock_inject):
        node = {
            'properties': {
                'inject_files': [
                    {'path': '/etc/default/grub', 'content': 'abcdef'},
                    {'path': '/etc/default/bluetooth', 'deleted': True},
                ]
            }
        }
        files = [
            {'path': '/boot/special.conf',
             'content': 'http://example.com/data',
             'mode': 0o600, 'dirmode': 0o750, 'owner': 0, 'group': 0},
            {'path': 'service.conf', 'partition': '/dev/disk/by-label/OPT'},
        ]

        inject_files.inject_files(node, [mock.sentinel.port], files)

        mock_inject.assert_has_calls([
            mock.call(node, [mock.sentinel.port], fl, '/dev/root', mock.ANY)
            for fl in node['properties']['inject_files'] + files
        ])
        http_get = mock_inject.call_args_list[0][0][4]
        self.assertTrue(http_get.verify)
        self.assertIsNone(http_get.cert)

    def test_verify_false(self, mock_inject):
        node = {
            'properties': {
                'inject_files': [
                    {'path': '/etc/default/grub', 'content': 'abcdef'},
                    {'path': '/etc/default/bluetooth', 'deleted': True},
                ]
            }
        }
        files = [
            {'path': '/boot/special.conf',
             'content': 'http://example.com/data',
             'mode': 0o600, 'dirmode': 0o750, 'owner': 0, 'group': 0},
            {'path': 'service.conf', 'partition': '/dev/disk/by-label/OPT'},
        ]

        inject_files.inject_files(node, [mock.sentinel.port], files, False)

        mock_inject.assert_has_calls([
            mock.call(node, [mock.sentinel.port], fl, '/dev/root', mock.ANY)
            for fl in node['properties']['inject_files'] + files
        ])
        http_get = mock_inject.call_args_list[0][0][4]
        self.assertFalse(http_get.verify)
        self.assertIsNone(http_get.cert)

    def test_invalid_type_on_node(self, mock_inject):
        node = {
            'properties': {
                'inject_files': 42
            }
        }
        self.assertRaises(errors.InvalidCommandParamsError,
                          inject_files.inject_files, node, [], [])
        mock_inject.assert_not_called()

    def test_invalid_type_in_param(self, mock_inject):
        node = {
            'properties': {}
        }
        self.assertRaises(errors.InvalidCommandParamsError,
                          inject_files.inject_files, node, [], 42)
        mock_inject.assert_not_called()


class TestValidateFiles(base.IronicAgentTest):

    def test_missing_path(self):
        fl = {'deleted': True}
        self.assertRaisesRegex(errors.InvalidCommandParamsError, 'path',
                               inject_files._validate_files, [fl], [])

    def test_unknown_fields(self):
        fl = {'path': '/etc/passwd', 'cat': 'meow'}
        self.assertRaisesRegex(errors.InvalidCommandParamsError, 'cat',
                               inject_files._validate_files, [fl], [])

    def test_root_without_on(self):
        fl = {'path': '/something', 'content': 'abcd'}
        self.assertRaisesRegex(errors.InvalidCommandParamsError, 'partition',
                               inject_files._validate_files, [fl], [])

    def test_no_directories(self):
        fl = {'path': '/something/else/', 'content': 'abcd'}
        self.assertRaisesRegex(errors.InvalidCommandParamsError, 'directories',
                               inject_files._validate_files, [fl], [])

    def test_content_and_deleted(self):
        fl = {'path': '/etc/password', 'content': 'abcd', 'deleted': True}
        self.assertRaisesRegex(errors.InvalidCommandParamsError,
                               'content .* with deleted',
                               inject_files._validate_files, [fl], [])

    def test_numeric_fields(self):
        for field in ('owner', 'group', 'mode', 'dirmode'):
            fl = {'path': '/etc/password', 'content': 'abcd', field: 'name'}
            self.assertRaisesRegex(errors.InvalidCommandParamsError,
                                   'must be a number',
                                   inject_files._validate_files, [fl], [])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import crypt

from oslo_log import log

from ironic_python_agent.extensions import base

LOG = log.getLogger()

PASSWORD_FILE = '/etc/ipa-rescue-config/ipa-rescue-password'


class RescueExtension(base.BaseAgentExtension):

    def write_rescue_password(self, rescue_password="", hashed=False):
        """Write rescue password to a file for use after IPA exits.

        :param rescue_password: Rescue password.
        :param hashed: Boolean default False indicating if the password
                       being provided is hashed or not. This will be changed
                       in a future version of ironic.
        """
        # DEPRECATED(TheJulia): In a future version of the ramdisk, we need
        # change the default such that a password is the default and that
        # if it is not, then the operation fails. Providing a default and
        # an override now that matches the present state allows us to
        # maintain our n-1, n, and n+1 theoretical support. Change
        # in the V or W cycles.
        LOG.debug('Writing hashed rescue password to %s', PASSWORD_FILE)
        password = str(rescue_password)
        hashed_password = None
        if hashed:
            hashed_password = password
        else:
            hashed_password = crypt.crypt(rescue_password)
        try:
            with open(PASSWORD_FILE, 'w') as f:
                f.write(hashed_password)
        except IOError as e:
            msg = ("Rescue Operation failed when writing the hashed rescue "
                   "password to the password file. Error %s") % e
            LOG.exception(msg)
            raise IOError(msg)

    @base.sync_command('finalize_rescue')
    def finalize_rescue(self, rescue_password="", hashed=False):
        """Sets the rescue password for the rescue user."""
        self.write_rescue_password(rescue_password, hashed)
        # IPA will terminate after the result of finalize_rescue is returned to
        # ironic to avoid exposing the IPA API to a tenant or public network
        self.agent.serve_api = False
        return



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic_python_agent.extensions import base
from ironic_python_agent import utils


class LogExtension(base.BaseAgentExtension):

    @base.sync_command('collect_system_logs')
    def collect_system_logs(self):
        """Collect system logs.

        Collect and package diagnostic and support data from the ramdisk.

        :raises: CommandExecutionError if failed to collect the system logs.
        :returns: A dictionary with the key `system_logs` and the value
                  of a gzipped and base64 encoded string of the file with
                  the logs.
        """
        logs = utils.collect_system_logs()
        return {'system_logs': logs}
